'''The :mod:`pulsar.apps.wsgi.content` introduces several utility classes
handling asynchronous content on a WSGI server. The main class of this module
is the :class:`AsyncString`. Web framework are smart ways of concatenating 
bytes which need to be displayed on a web page or returned as a response to
and HTTP client.

Asynchronous String
=====================

.. autoclass:: AsyncString
   :members:
   :member-order: bysource

Asynchronous Json
=====================

.. autoclass:: Json
   :members:
   :member-order: bysource
   
.. _wsgi-html:
   
Asynchronous Html
=====================

.. autoclass:: Html
   :members:
   :member-order: bysource
   
.. _wsgi-html-document:

Html Document
==================

Document
~~~~~~~~~~

.. autoclass:: HtmlDocument
   :members:
   :member-order: bysource
   
Head
~~~~~~~~~~

.. autoclass:: Head
   :members:
   :member-order: bysource
   
Body
~~~~~~~~~~

.. autoclass:: Body
   :members:
   :member-order: bysource
   
StreamRenderer
==================

.. autoclass:: StreamRenderer
   :members:
   :member-order: bysource
   
Html Factory
=================

.. autofunction:: html_factory


'''
import json
from collections import Mapping
from functools import partial
from copy import copy

from pulsar import Deferred, multi_async, is_async, maybe_async, is_failure, async
from pulsar.utils.pep import iteritems, is_string, ispy3k
from pulsar.utils.structures import AttributeDictionary, recursive_update,\
                                    OrderedDict
from pulsar.utils.html import slugify, INLINE_TAGS, tag_attributes, attr_iter,\
                                csslink, dump_data_value, child_tag
from pulsar.utils.httpurl import remove_double_slash, urljoin

from .html import html_visitor

__all__ = ['AsyncString', 'Html',
           'Json', 'HtmlDocument',
           'html_factory', 'Media']


class StreamRenderer(Deferred):
    '''A specialised :class:`pulsar.Deferred` returned by the
:meth:`AsyncString.content` method.'''
    def __init__(self, stream, renderer, handle_value=None, **params):
        super(StreamRenderer, self).__init__()
        handle_value = handle_value or self._handle_value
        self._m = multi_async(stream, raise_on_error=True,
                              handle_value=handle_value, **params)
        self._m.add_callback(renderer).add_both(self.callback)

    def _handle_value(self, value):
        '''It makes sure that :class:`Content` is unwond.
Ideally this should not occur since the request object is not available.
Users should always call the content method before.'''
        if isinstance(value, AsyncString):
            return value.content()
        else:
            return value
    
if ispy3k:
    def stream_to_string(stream):
        for value in stream:
            if value is None:
                continue
            elif isinstance(value, bytes):
                yield value.decode('utf-8')
            elif isinstance(value, str):
                yield value
            else:
                yield str(value)

else: #pragma nocover
    def stream_to_string(stream):
        for value in stream:
            if value is None:
                continue
            elif isinstance(value, unicode):
                yield value
            else:
                yield str(value)
    

class AsyncString(object):
    '''Class for asynchronous strings which can be used with
pulsar WSGI servers.
'''
    content_type = None
    '''Content type for this :class:`AsyncString`'''
    encoding = None
    '''Charset encoding for this :class:`AsyncString`'''
    
    def __init__(self, *children):
        self._streamed = False
        self._children = []
        self._parent = None
        for child in children:
            self.append(child)
    
    @property
    def parent(self):
        '''The :class:`AsyncString` element which contains this
:class:`AsyncString`.'''
        return self._parent
    
    @property
    def children(self):
        '''A copy of all children of this :class:`AsyncString`. Children can
be other :class:`AsyncString` or string or bytes, depending on implementation.
:attr:`children` are added and removed via the :meth:`append` and
:meth:`remove` methods.'''
        return copy(self._children)
    
    def __repr__(self):
        return self.__class__.__name__()
    
    def __str__(self):
        return self.__repr__()
    
    def append(self, child):
        '''Append ``child`` to the list of :attr:`children` of this
:class:`AsyncString`.

:param child: String, bytes or another :class:`AsyncString`. If it is an
    :class:`AsyncString`, this instance will be set as its :attr:`parent`.
    If ``child`` is ``None``, this method does nothing.
    
'''
        # make sure that child is not in child
        if child is not None:
            if isinstance(child, AsyncString):
                child_parent = child._parent
                if self._parent is child:
                    # the parent is the child we are appending, set the parent
                    # to be the parent of child
                    self._parent = child_parent
                if child_parent:
                    # remove child from the child parent
                    child_parent.remove(child)
                child._parent = self
            self._children.append(child)
        
    def remove(self, child):
        '''Remove a ``child`` from the list of :attr:``children``.'''
        try:
            self._children.remove(child)
            if isinstance(child, AsyncString):
                child._parent = None
        except ValueError:
            pass
    
    def remove_all(self):
        '''Remove all :attr:`children`.'''
        for child in self._children:
            if isinstance(child, AsyncString):
                child._parent = None
        self._children = []
        
    def append_to(self, parent):
        '''Append itself to ``parent``. Return ``self``.'''
        parent.append(self)
        return self
    
    def content(self, request=None):
        '''Return the :class:`StreamRenderer` for this instance.
This method can be called once only since it invokes the :meth:`stream`
method.'''
        res = self.stream(request)
        if is_async(res):
            return res.add_callback(lambda r: StreamRenderer(r, self.to_string))
        else:
            return StreamRenderer(res, self.to_string)
    
    def stream(self, request):
        '''This is the most important method of an :class:`AsyncString`.
It is called by :meth:`content` or by the :attr:`parent` of this
:class:`AsyncString`. It returns an iterable (list, tuple, generator) over
strings, that means ``unicode/str`` for python 2, and ``str`` for python 3 or
:ref:`asynchronous elements <tutorials-coroutine>` which result in strings.
This method can be called once only, otherwise a
:class:`RuntimeError` occurs.'''
        if self._streamed:
            raise RuntimeError('%s already streamed' % self)
        if request:
            self.stream_started(request)
        self._streamed = True
        return self.do_stream(request)
    
    def stream_started(self, request):
        '''Hook called if ``request`` is available, just before the
:meth:`do_stream` is executed.'''
        pass
    
    @async()
    def http_response(self, request):
        '''Return a, possibly, :ref:`asynchronous WSGI iterable <wsgi-async>`.
This method asynchronously wait for :meth:`content` and subsequently
starts the wsgi response.'''
        response = request.response
        response.content_type = self.content_type
        body = yield self.content(request)
        response.content = body
        yield response
        
    def do_stream(self, request):
        '''Perform the actual streaming of this :class:`AsyncString`.
This method can be re-implemented by subclasses.'''
        for child in self._children:
            if isinstance(child, AsyncString):
                for bit in child.stream(request):
                    yield bit
            else:
                yield child
                
    def to_string(self, stream):
        '''Once the :class:`StreamRenderer`, returned by :meth:`content`
method, is ready, meaning it has no more
asynchronous elements, this method get called to transform the stream into the
content string. This method can be overwritten by derived classes.

:param stream: a collections containing ``strings/bytes`` used to build the
    final ``string/bytes``.
:return: a string or bytes
'''
        return ''.join(stream_to_string(stream))
    
    def render(self, request=None):
        '''A shortcut function for synchronously rendering a Content.
This is useful during testing. It is the synchronous equivalent of
:meth:`content`.'''
        value = maybe_async(self.content(request))
        if is_failure(value):
            value.raise_all()
        elif is_async(value):
            raise ValueError('Could not render. Asynchronous value')
        else:
            return value
            

class Json(AsyncString):
    '''An :class:`AsyncString` which renders into a json string.
The :attr:`AsyncString.content_type` attribute is set to
``application/json``.'''
    def __init__(self, *children, **params):
        super(Json, self).__init__(*children)
        self.parameters = AttributeDictionary(params)
        
    @property
    def json(self):
        '''The ``json`` encoder/decoder handler. If a ``json`` entry is not
provided during initialisation, the standard python ``json`` module
is used.'''
        return self.parameters.json or json
        
    @property
    def content_type(self):
        return 'application/json'
        
    def to_string(self, stream):
        if len(stream) == 1:
            return self.json.dumps(stream[0])
        else:
            return self.json.dumps(stream)
        

def html_factory(tag, **defaults):
    '''Returns an :class:`Html` factory function for ``tag`` and a given
dictionary of ``defaults`` parameters. For example::

    >>> input_factory = html_factory('input', type='text')
    >>> html = input_factory(value='bla')
    
    '''
    def html_input(*children, **params):
        p = defaults.copy()
        p.update(params)
        return Html(tag, *children, **p)
    return html_input
    
    
class Html(AsyncString):
    '''An :class:`AsyncString` for html elements.
The :attr:`AsyncString.content_type` attribute is set to ``text/html``.

:param tag: Set the :attr:`tag` attribute. Must be given and can be ``None``.
:param children: Optional children which will be added via the
    :meth:`AsyncString.append` method.
:param params: Optional keyed-value parameters.

Special (optional) parameters:

* ``cn`` class name or list of class names.
* ``attr`` dictionary of attributes to add.
* ``data`` dictionary of data to add (rendered as HTML data).
* ``type`` type of element, only supported for tags which accept the ``type``
  attribute (for example the ``input`` tag).

Any other keyed-value parameter will be added as attribute, if in the set of
:attr:`available_attributes` or as :meth:`data`.
'''
    def __init__(self, tag, *children, **params):
        self._tag = tag
        self._classes = set()
        self._data = {}
        self._attr = {}
        self._css = {}
        self._setup(**params)
        super(Html, self).__init__(*children)
        
    @property
    def content_type(self):
        return 'text/html'
    
    @property
    def classes(self):
        '''Set of classes for this :class:`Html` element.'''
        return self._classes
    
    @property
    def tag(self):
        '''The tag for this HTML element, ``div``, ``a``, ``table`` and so
forth. It can be ``None``.'''
        return self._tag
    
    @property
    def available_attributes(self):
        '''The list of valid HTML attributes for this :attr:`tag`.'''
        return tag_attributes(self._tag, self._attr.get('type'))
    
    def get_form_value(self):
        '''Return the value of this :class:`Html` element when it is contained
in a Html form element. For most element it gets the ``value`` attribute.'''
        return self._visitor.get_form_value(self)
    
    def set_form_value(self, value):
        '''Set the value of this :class:`Html` element when it is contained
in a Html form element. For most element it sets the ``value`` attribute.'''
        self._visitor.set_form_value(self, value)
        
    def __repr__(self):
        if self._tag and self._tag in INLINE_TAGS:
            return '<%s%s/>' % (self._tag, self.flatatt())
        elif self._tag:
            return '<%s%s>' % (self._tag, self.flatatt())
        else:
            return self.__class__.__name__
    
    def append(self, child):
        if child:
            tag = child_tag(self._tag)
            if tag:
                if isinstance(child, Html):
                    if child.tag != tag:
                        child = Html(tag, child)
                elif not child.startswith('<%s' % tag):
                    child = Html(tag, child)
        super(Html, self).append(child)
    
    def _setup(self, cn=None, attr=None, css=None, data=None, type=None,
               **params):
        self._visitor = html_visitor(self._tag)
        self.addClass(cn)
        self.data(data)
        self.attr(attr)
        self.css(css)
        attributes = self.available_attributes
        if type and 'type' in attributes:
            self.attr('type', type)
            attributes = self.available_attributes
        for name, value in iteritems(params):
            if name in attributes:
                self.attr(name, value)
            else:
                self.data(name, value)
        
    def attr(self, *args):
        '''Add the specific attribute to the attribute dictionary
with key ``name`` and value ``value`` and return ``self``.'''
        if not args:
            return self._attr
        result, adding = self._attrdata(self._attr, *args)
        if adding:
            available_attributes = self.available_attributes
            for name, value in iteritems(result):
                if value is not None:
                    if name in available_attributes:
                        self._attr[name] = value
                    elif name is 'value':
                        self.append(value)
            result = self
        return result
    
    def data(self, *args):
        '''Add or retrieve data values for this :class:`Html`.'''
        if not args:
            return self._data
        result, adding = self._attrdata(self._data, *args)
        if adding:
            recursive_update(self._data, result)
            result = self
        return result
    
    def addClass(self, cn):
        '''Add the specific class names to the class set and return ``self``.'''
        if cn:
            if isinstance(cn,(tuple,list,set,frozenset)):
                add = self.addClass
                for c in cn:
                    add(c)
            else:
                add = self._classes.add
                for cn in cn.split():
                    add(slugify(cn, rtx='-'))
        return self

    def hasClass(self, cn):
        '''``True`` if ``cn`` is a class of self.'''
        return cn in self._classes
    
    def removeClass(self, cn):
        '''Remove classes'''
        if cn:
            ks = self._classes
            for cn in cn.split():
                if cn in ks:
                    ks.remove(cn)
        return self
    
    def flatatt(self, **attr):
        '''Return a string with atributes to add to the tag'''
        cs = ''
        attr = self._attr.copy()
        if self._classes:
            cs = ' '.join(self._classes)
            attr['class'] = cs
        if self._css:
            attr['style'] = ' '.join(('%s:%s;' % (k,v)\
                                       for k,v in self._css.items()))
        for k, v in self._data.items():
            attr['data-%s' % k] = dump_data_value(v)
        if attr:
            return ''.join(attr_iter(attr))
        else:
            return ''

    def css(self, mapping=None):
        '''Update the css dictionary if *mapping* is a dictionary, otherwise
 return the css value at *mapping*.'''
        if mapping is None:
            return self._css
        elif isinstance(mapping, Mapping):
            self._css.update(mapping)
            return self
        else:
            return self._css.get(mapping)
    
    def hide(self):
        '''Same as jQuery hide method.'''
        self._css.update({'display': 'none'})
        return self
    
    def show(self):
        '''Same as jQuery show method.'''
        self._css.pop('display', None)
        return self
    
    def do_stream(self, request):
        if self._tag and self._tag in INLINE_TAGS:
            yield '<%s%s>' % (self._tag, self.flatatt())
        else:
            if self._tag:
                yield '<%s%s>' % (self._tag, self.flatatt())
            for child in self._children:
                if isinstance(child, AsyncString):
                    for bit in child.stream(request):
                        yield bit
                else:
                    yield child
            if self._tag:
                yield '</%s>' % self._tag
    
    def _attrdata(self, cont, name, *val):
        if isinstance(name, Mapping):
            if val:
                raise TypeError('Cannot set a value to %s' % name)
            return name, True
        else:
            if val:
                if len(val) == 1:
                    return {name: val[0]}, True
                else:
                    raise TypeError('Too may arguments')
            else:
                return cont.get(name), False
            

class Media(AsyncString):
    
    def __init__(self, media_path):
        super(Media, self).__init__()
        self.media_path = media_path
        self._children = OrderedDict()
    
    def append(self, value):
        raise NotImplementedError
        
    def absolute_path(self, path):
        if path.startswith('http://') or path.startswith('https://')\
            or path.startswith('/'):
            return path
        return remove_double_slash('/%s/%s' % (self.media_path, path))


class Css(Media):
        
    def append(self, value):
        if value:
            if isinstance(value, str):
                value = {'all': [value]}
            for media, values in value.items():
                m = self._children.get(media, [])
                for value in values:
                    if value not in m:
                        if not isinstance(value, (tuple, list)):
                            value = csslink(value, None)
                        else:
                            value = csslink(*value)
                        m.append(value)
                self._children[media] = m

    def do_stream(self, request):
        for medium in sorted(self._children):
            paths = self._children[medium]
            medium = '' if medium == 'all' else " media='%s'" % medium
            for path in paths:
                link = "<link href='%s' type='text/css'%s rel='stylesheet'/>\n"\
                        % (self.absolute_path(path.link), medium)
                if path.condition:
                    link = '<!--[if %s]>%s<![endif]-->' % (path.condition, link)
                yield link
    
    
class Scripts(Media):
        
    def append(self, child):
        if child and is_string(child):
            path = self.absolute_path(child)
            script = Html('script', src=path, type='text/javascript')
            self._children[script] = script
            return script
    
    def do_stream(self, request):
        for child in self._children.values():
            for bit in child.stream(request):
                yield bit
            yield '\n'
        
        
class Head(Html):
    ''':class:`Html` head tag. It contains :class:`Html` handlers for the
various part of an HTML Head element.
    
.. attribute:: title

    Text in the title tag
    
.. attribute:: meta

    A container of :class:`Html` meta tags. To add new meta tags use the
    :meth:`add_meta` method rather than accessing the :attr:`meta`
    attribute directly.
    
.. attribute:: scripts

    A container of Javascript files to render at the end of the body tag.
    To add new javascript files simply use the append method on
    this attribute. You can add relative paths::
    
        html.head.scripts.append('/media/js/scripts.js')
    
    as well as absolute paths::
    
        html.head.scripts.append('https://ajax.googleapis.com/ajax/libs/jquery/1.7.2/jquery.min.js')

'''
    def __init__(self, media_path=None, title=None, meta=None, charset=None):
        super(Head, self).__init__('head')
        self.title = title
        self.append(Html(None, meta))
        self.append(Css(media_path))
        self.append(Scripts(media_path))
        self.add_meta(charset=charset or 'utf-8')
    
    @property
    def meta(self):
        return self._children[0]
    
    @property
    def links(self):
        return self._children[1]
    
    def __get_scripts(self):
        return self._children[2]
    def __set_scripts(self, scripts):
        self._children[2] = scripts
    scripts = property(__get_scripts, __set_scripts)
    
    def do_stream(self, request):
        if self.title:
            self._children.insert(0, '<title>%s</title>' % self.title)
        return super(Head, self).do_stream(request)
        
    def body(self, request):
        return self.js_body.content(request)
    
    def add_meta(self, **kwargs):
        '''Add a new :class:`Html` meta tag to the :attr:`meta` collection.'''
        meta = Html('meta', **kwargs)
        self.meta.append(meta)
            
    def add(self, other):
        if isinstance(other, Head):
            self.style.append(other.style)
            self.meta.append(other.meta)
            self.js_head.append(other.js_head)
            self.js_body.append(other.js_body)
        return self

    def __add__(self, other):
        if isinstance(other, Media):
            return Media(media=self).add(other)
        else:
            return self
        
        
class Body(Html):
    ''':class:`Html` body tag.
    
.. attribute:: scripts

    A container of Javascript files to render at the end of the body tag.
    The usage is the same as :attr:`Head.scripts`.    
'''
    def __init__(self, media_path=None):
        super(Body, self).__init__('body')
        self.scripts = Scripts(media_path)
    
    def do_stream(self, request):
        '''Render the widget. It accept two optional parameters, a http
request object and a dictionary for rendering children with a key.

:parameter request: Optional request object.
'''
        self.append(self.scripts)
        return super(Body, self).do_stream(request)
                
                
class HtmlDocument(Html):
    '''HTML5 asynchronous document. An instance of this class can be obtained
via the :attr:`pulsar.apps.wsgi.wrappers.WsgiRequest.html_document` attribute.
    
.. attribute:: head

    The :class:`Head` part of this :class:`HtmlDocument`

.. attribute:: body

    The :class:`Body` part of this :class:`HtmlDocument`
    
'''
    def __init__(self, title=None, media_path='/media/', charset=None,
                 **params):
        super(HtmlDocument, self).__init__(None, **params)
        self.head = Head(title=title, media_path=media_path, charset=charset)
        self.body = Body(media_path=media_path)
    
    def __call__(self, title=None, body=None, media_path=None):
        if title:
            self.head.title = title
        if media_path:
            self.head.scripts.media_path = media_path
            self.head.links.media_path = media_path
            self.body.scripts.media_path = media_path
        self.body.append(body)
        return self
        
    @async()
    def do_stream(self, request):
        body = yield self.body.content(request)
        head = yield self.head.content(request)
        yield  ('<!DOCTYPE html>\n', '<html%s>\n' % self.flatatt(),
                head, '\n', body, '\n</html>')



'''
An implementation of a :class:`TaskBackend` which uses redis as data server.
Requires python-stdnet_

.. _python-stdnet: https://pypi.python.org/pypi/python-stdnet
'''
from stdnet import odm

from pulsar import async
from pulsar.apps.tasks import backends, states, LOGGER
from pulsar.utils.log import local_method


class TaskData(odm.StdModel):
    id = odm.SymbolField(primary_key=True)
    name = odm.SymbolField()
    status = odm.SymbolField()
    args = odm.PickleObjectField()
    kwargs = odm.PickleObjectField()
    result = odm.PickleObjectField()
    from_task = odm.SymbolField(required=False)
    time_executed = odm.DateTimeField(index=False)
    time_started = odm.DateTimeField(required=False, index=False)
    time_ended = odm.DateTimeField(required=False, index=False)
    expiry = odm.DateTimeField(required=False, index=False)
    meta = odm.JSONField()
    #
    # List where all TaskData objects are queued
    queue = odm.ListField(class_field=True)
    
    class Meta:
        app_label = 'tasks'
        
    def as_task(self):
        params = dict(self.meta or {})
        for field in self._meta.scalarfields:
            params[field.name] = getattr(self, field.attname, None)
        return backends.Task(self.id, **params)
    
    def __unicode__(self):
        return '%s (%s)' % (self.name, self.status)


class TaskBackend(backends.TaskBackend):
    
    @local_method
    def task_manager(self):
        self.local.models = odm.Router(self.connection_string)
        self.local.models.register(TaskData)
        return self.local.models.taskdata
    
    def num_tasks(self):
        '''Retrieve the number of tasks in the task queue.'''
        task_manager = self.task_manager()
        return task_manager.queue.size()
    
    @async()
    def put_task(self, task_id):
        task_data = yield self._get_task(task_id)
        if task_data:
            task_data.status = states.QUEUED
            task_data = yield task_data.save()
            yield self.task_manager().queue.push_back(task_data.id)
            yield task_data.id
    
    @async()    
    def save_task(self, task_id, **params):
        # Called by self when the task need to be saved
        task_manager = self.task_manager()
        task_data = yield self._get_task(task_id)
        if task_data:
            for field, value in params.items():
                if field in task_data._meta.dfields:
                    setattr(task_data, field, value)
                else:
                    # not a field put value in the meta json field
                    task_data.meta[field] = value
            yield task_data.save()
        else:
            task_data = yield task_manager.new(id=task_id, **params)
        task = task_data.as_task()
        yield task_id
    
    def get_task(self, task_id=None, timeout=1):
        task_manager = self.task_manager()
        #
        if not task_id:
            #pool = task_manager.backend.client.connection_pool
            #LOGGER.info('CONNECTIONS: AVAILABLE %s, CONCURRENT %s, TOTAL %s',
            #            pool.available_connections, pool.concurrent_connections,
            #            pool.available_connections+pool.concurrent_connections)
            task_id = yield task_manager.queue.block_pop_front(timeout=timeout)
        if task_id:
            task_data = yield self._get_task(task_id)
            if task_data:
                yield task_data.as_task()
        
    def get_tasks(self, **filters):
        task_manager = self.task_manager()
        tasks = yield task_manager.filter(**filters).all()
        yield [t.as_task() for t in tasks]
        
    def delete_tasks(self, ids=None):
        deleted = []
        if ids:
            task_manager = self.task_manager()
            tasks = yield task_manager.filter(id=ids).all()
            yield task_manager.filter(id=ids).delete()
            for task_data in tasks:
                deleted.append(task_data.id)
        yield deleted
        
    ############################################################################
    ##    INTERNALS
    def _get_task(self, task_id):
        tasks = yield self.task_manager().filter(id=task_id).all()
        if tasks:
            yield tasks[0]
            


'''\
Pulsar ships with an asynchronous :class:`TaskQueue` built on top
:ref:`pulsar application framework <apps-framework>`. Task queues are used
as a mechanism to distribute work across threads/processes or machines.
Pulsar :class:`TaskQueue` is highly customizable, it can run in multi-threading
or multiprocessing (default) mode and can share :class:`backends.Task` across
several machines.
By creating :class:`models.Job` classes in a similar way you do for celery_,
this application gives you all you need for running them with very
little setup effort::

    from pulsar.apps import tasks

    tq = tasks.TaskQueue(tasks_path=['path.to.tasks.*'])
    tq.start()
    
To get started, follow the these points:

* Create the script which runs your application, in the
  :ref:`taskqueue tutorial <tutorials-taskqueue>` the script is called
  ``manage.py``.
* Create the modules where :ref:`jobs <app-taskqueue-job>` are implemented. It
  can be a directory containing several submodules as explained in the
  :ref:`task paths parameter <app-tasks_path>`.
  

.. _app-taskqueue-job:

Configuration
~~~~~~~~~~~~~~~~
A :class:`TaskQueue` accepts several configuration parameters on top of the
standard :ref:`application settings <settings>`:

.. _app-tasks_path:

* The :ref:`task_paths <setting-task_paths>` parameter specify
  a list of python paths where to collect :class:`models.Job` classes::
  
      task_paths = ['myjobs','another.moduledir.*']
      
  The ``*`` at the end of the second module indicates to collect
  :class:`models.Job` from all submodules of ``another.moduledir``.
  
* The :ref:`schedule_periodic <setting-schedule_periodic>` flag indicates
  if the :class:`TaskQueue` can schedule :class:`models.PeriodicJob`. Usually,
  only one running :class:`TaskQueue` application is responsible for
  scheduling tasks while all the other, simply consume tasks.
  This parameter can also be specified in the command line via the
  ``--schedule-periodic`` flag. Default: ``False``.
  
* The :ref:`task_backend <setting-task_backend>` parameter is a url
  type string which specifies the :class:`backends.TaskBackend`
  to use.


.. _app-taskqueue-app:

Task queue application
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

.. autoclass:: TaskQueue
   :members:
   :member-order: bysource
   

.. _celery: http://celeryproject.org/
'''
import os
from datetime import datetime

import pulsar
from pulsar import to_string, command
from pulsar.utils.log import local_property
from pulsar.utils.config import section_docs

from .models import *
from .states import *
from .backends import *
from .rpc import *


section_docs['Task Consumer'] = '''
This section covers configuration parameters used by CPU bound type applications
such as the :ref:`distributed task queue <apps-taskqueue>` and the
:ref:`test suite <apps-test>`.'''


class TaskSetting(pulsar.Setting):
    virtual = True
    app = 'tasks'
    section = "Task Consumer"


class ConcurrentTasks(TaskSetting):
    name = "concurrent_tasks"
    flags = ["--concurrent-tasks"]
    validator = pulsar.validate_pos_int
    type = int
    default = 5
    desc = """\
        The maximum number of concurrent tasks for a worker.
        
        When a task worker reach this number it stops polling for more tasks
        until one or more task finish. It should only affect task queues under
        significant load.
        Must be a positive integer. Generally set in the range of 5-10.
        """

class TaskBackendConnection(TaskSetting):
    name = "task_backend"
    flags = ["--task-backend"]
    default = "local://"
    desc = '''\
        Task backend.

        A task backend is string which connect to the backend storing Tasks)
        which accepts one parameter only and returns an instance of a
        distributed queue which has the same API as
        :class:`pulsar.MessageQueue`. The only parameter passed to the
        task queue factory is a :class:`pulsar.utils.config.Config` instance.
        This parameters is used by :class:`pulsar.apps.tasks.TaskQueue`
        application.'''

    
class TaskPaths(TaskSetting):
    name = "task_paths"
    validator = pulsar.validate_list
    default = []
    desc = """\
        List of python dotted paths where tasks are located.
        
        This parameter can only be specified during initialization or in a
        :ref:`config file <setting-config>`.
        """
        
        
class SchedulePeriodic(TaskSetting):
    name = 'schedule_periodic'
    flags = ["--schedule-periodic"]
    validator = pulsar.validate_bool
    action = "store_true"
    default = False
    desc = '''\
        Enable scheduling of periodic tasks.
        
        If enabled, :class:`pulsar.apps.tasks.PeriodicJob` will produce
        tasks according to their schedule.
        '''


class TaskQueue(pulsar.Application):
    '''A :class:`pulsar.apps.Application` for consuming
task.Tasks and managing scheduling of tasks via a
:class:`scheduler.Scheduler`.'''
    backend = None
    name = 'tasks'
    cfg = pulsar.Config(apps=('tasks',), timeout=600)

    def request_instance(self, request):
        return self.scheduler.get_task(request)
    
    def monitor_start(self, monitor):
        '''When the monitor starts create the :class:`backends.TaskBackend`.'''
        if self.callable:
            self.callable()
        self.backend = TaskBackend.make(
                                self.cfg.task_backend,
                                name=self.name,
                                task_paths=self.cfg.task_paths,
                                schedule_periodic=self.cfg.schedule_periodic,
                                max_tasks=self.cfg.max_requests,
                                backlog=self.cfg.concurrent_tasks)
        
    def monitor_task(self, monitor):
        '''Override the :meth:`pulsar.apps.Application.monitor_task` callback
to check if the :attr:`scheduler` needs to perform a new run.'''
        super(TaskQueue, self).monitor_task(monitor)
        if self.backend and monitor.running:
            if self.backend.next_run <= datetime.now():
                self.backend.tick()

    def worker_start(self, worker):
        self.backend.start(worker)
        
    def worker_stopping(self, worker):
        self.backend.close(worker)
        
    def actorparams(self, monitor, params):
        # Make sure we invoke super function so that we get the distributed
        # task queue
        params = super(TaskQueue, self).actorparams(monitor, params)
        # workers do not schedule periodic tasks
        params['app'].cfg.set('schedule_periodic', False)
        return params
    
    def worker_info(self, worker, data):
        be = self.backend
        tasks = {'concurrent': list(be.concurrent_tasks),
                 'processed': be.processed}
        data['tasks'] = tasks
        return data
     
    
@command()
def next_scheduled(request, jobnames=None):
    actor = request.actor
    return actor.app.backend.next_scheduled(jobnames)


'''
Pulsar concurrent framework and applications can be used with twisted_, an
event driven network engine for python. Twisted has implementation
for several protocols which can be used in pulsar by importing the
:mod:`pulsar.apps.tx` module::

    from pulsar.apps.tx import twisted

The implementation replaces the twisted reactor with a proxy for
:class:`pulsar.EventLoop`.
Twisted Deferred and Failures are made compatible with pulsar
by installing two different discovery functions via the :func:`pulsar.set_async`
function.

Threads, signal handling, scheduling and so forth is handled by pulsar itself,
twisted implementation is switched off.

Pulsar Reactor
====================

.. autoclass:: PulsarReactor
   :members:
   :member-order: bysource
   

.. _twisted: http://twistedmatrix.com/
'''
import twisted
from twisted.internet.main import installReactor
from twisted.internet.posixbase import PosixReactorBase
from twisted.internet.defer import Deferred
from twisted.python.failure import Failure

import pulsar
from pulsar.async.defer import default_maybe_async, default_maybe_failure, set_async
from pulsar.utils.pep import get_event_loop

    
def _maybe_async(obj, **params):
    if isinstance(obj, Deferred):
        d = pulsar.Deferred()
        d._twisted_deferred = obj
        obj.addBoth(d.callback)
        obj = d
    return default_maybe_async(obj, **params)

def _maybe_failure(e, msg=''):
    if isinstance(e, Failure):
        return pulsar.Failure((e.type, e.value, e.tb), msg=msg)
    else:
        return default_maybe_failure(e, msg=msg)
        

# Set the new async discovery functions
set_async(_maybe_async, _maybe_failure)


class PulsarReactor(PosixReactorBase):
    '''A twisted reactor which acts as a proxy to pulsar eventloop.'''
    def callLater(self, _seconds, f, *args, **kw):
        return get_event_loop().call_later(_seconds, lambda : f(*args, **kw))
    
    def callFromThread(self, f, *args, **kw):
        return get_event_loop().call_soon_threadsafe(lambda : f(*args, **kw))
    
    def callInThread(self, f, *args, **kw):
        return get_event_loop().call_soon(lambda : f(*args, **kw))
        
    def addReader(self, reader):
        return get_event_loop().add_reader(reader.fileno(), reader.doRead)
        
    def addWriter(self, writer):
        return get_event_loop().add_writer(writer.fileno(), writer.doWrite)
        
    def removeReader(self, reader):
        return get_event_loop().remove_reader(reader.fileno())
        
    def removeWriter(self, writer):
        return get_event_loop().remove_writer(writer.fileno())
        
    def removeAll(self):
        return get_event_loop().remove_all()
    
    ############################################################################
    # Functions not needed by pulsar
    _registerAsIOThread = False
    
    def installWaker(self):
        pass
    
    def mainLoop(self):
        pass
    
    def doIteration(self, delay):
        pass
    
    def spawnProcess(self, *args, **kwargs):
        raise NotImplementedError('Cannot spawn process from Pulsar reactor')
    
    def _initThreads(self):
        pass
    
    def _handleSignals(self):
        pass
        
    
_reactor = PulsarReactor()
installReactor(_reactor)
_reactor.run()


'''This is a substantial module which imporfts several classes and functions
from the standard library in a python 2.6 to python 3.3 compatible fashion.
On top of that, it implements the :class:`HttpClient` for handling synchronous
and asynchronous HTTP requests in a pythonic way.

It is a thin layer on top of urllib2 in python2 / urllib in Python 3.
Several opensource efforts have been used as source of snippets:

* http-parser_
* request_
* urllib3_
* werkzeug_


.. _tools-http-headers:

HTTP Headers
~~~~~~~~~~~~~~~~~

.. autoclass:: Headers
   :members:
   :member-order: bysource
   
   
.. _tools-http-parser:

HTTP Parser
~~~~~~~~~~~~~~~~~

.. autoclass:: HttpParser
   :members:
   :member-order: bysource

parse cookie
~~~~~~~~~~~~~~~~~~~~~

.. autofunction:: parse_cookie

.. _http-parser: https://github.com/benoitc/http-parser
.. _urllib3: https://github.com/shazow/urllib3
.. _request: https://github.com/kennethreitz/requests
.. _werkzeug: https://github.com/mitsuhiko/werkzeug
.. _`HTTP cookie`: http://en.wikipedia.org/wiki/HTTP_cookie
'''
import os
import sys
import re
import io
import string
import time
import mimetypes
import platform
import codecs
import socket
import logging
from functools import reduce
from hashlib import sha1, md5
from uuid import uuid4
from datetime import datetime, timedelta
from email.utils import formatdate
from io import BytesIO
import zlib
from collections import deque
from copy import copy

try:
    from http_parser.parser import HttpParser as CHttpParser
    hasextensions = True
    _Http_Parser = CHttpParser
except ImportError: #pragma    nocover
    hasextensions = False
    _Http_Parser = None

def setDefaultHttpParser(parser):
    global _Http_Parser
    _Http_Parser = parser
    
def http_parser(**kwargs):
    global _Http_Parser
    return _Http_Parser(**kwargs)

from .structures import mapping_iterator, OrderedDict
from .pep import ispy3k, ispy26

create_connection = socket.create_connection
LOGGER = logging.getLogger('httpurl')

try:
    from select import poll, POLLIN
except ImportError: #pragma    nocover
    poll = False
    try:
        from select import select
    except ImportError: #pragma    nocover
        select = False
        
try:    # Compiled with SSL?
    BaseSSLError = None
    ssl = None
    import ssl
    BaseSSLError = ssl.SSLError
except (ImportError, AttributeError):   # pragma : no cover 
    pass

if ispy3k: # Python 3
    from urllib import request as urllibr
    from http import client as httpclient
    from urllib.parse import quote, unquote, urlencode, urlparse, urlsplit,\
                             parse_qs, parse_qsl, splitport, urlunparse,\
                             urljoin
    from http.client import responses
    from http.cookiejar import CookieJar, Cookie
    from http.cookies import SimpleCookie, BaseCookie, Morsel, CookieError

    string_type = str
    getproxies_environment = urllibr.getproxies_environment
    ascii_letters = string.ascii_letters
    zip = zip
    map = map
    range = range
    chr = chr
    iteritems = lambda d : d.items()
    itervalues = lambda d : d.values()
    is_string = lambda s: isinstance(s, str)
    is_string_or_native_string = is_string

    def to_bytes(s, encoding=None, errors=None):
        errors = errors or 'strict'
        encoding = encoding or 'utf-8'
        if isinstance(s, bytes):
            if encoding != 'utf-8':
                return s.decode('utf-8', errors).encode(encoding, errors)
            else:
                return s
        else:
            return ('%s'%s).encode(encoding, errors)

    def to_string(s, encoding=None, errors='strict'):
        """Inverse of to_bytes"""
        if isinstance(s, bytes):
            return s.decode(encoding or 'utf-8', errors)
        else:
            return str(s)

    def native_str(s, encoding=None):
        if isinstance(s, bytes):
            return s.decode(encoding or 'utf-8')
        else:
            return s

    def force_native_str(s, encoding=None):
        if isinstance(s, bytes):
            return s.decode(encoding or 'utf-8')
        elif not isinstance(s, str):
            return str(s)
        else:
            return s

else:   # pragma : no cover
    import urllib2 as urllibr
    import httplib as httpclient
    from urllib import quote, unquote, urlencode, getproxies_environment,\
                        splitport
    from urlparse import urlparse, urlsplit, parse_qs, urlunparse, urljoin,\
                         parse_qsl
    from httplib import responses
    from cookielib import CookieJar, Cookie
    from Cookie import SimpleCookie, BaseCookie, Morsel, CookieError
    from itertools import izip as zip, imap as map

    string_type = unicode
    ascii_letters = string.letters
    range = xrange
    chr = unichr
    iteritems = lambda d : d.iteritems()
    itervalues = lambda d : d.itervalues()
    is_string = lambda s: isinstance(s, unicode)
    is_string_or_native_string = lambda s: isinstance(s, basestring)

    if sys.version_info < (2, 7):
        #
        def create_connection(address, timeout=socket._GLOBAL_DEFAULT_TIMEOUT,
                              source_address=None):
            """Form Python 2.7"""
            host, port = address
            err = None
            for res in socket.getaddrinfo(host, port, 0, socket.SOCK_STREAM):
                af, socktype, proto, canonname, sa = res
                sock = None
                try:
                    sock = socket.socket(af, socktype, proto)
                    if timeout is not socket._GLOBAL_DEFAULT_TIMEOUT:
                        sock.settimeout(timeout)
                    if source_address:
                        sock.bind(source_address)
                    sock.connect(sa)
                    return sock
                except Exception as _:
                    err = _
                    if sock is not None:
                        sock.close()
            if err is not None:
                raise err
            else:
                raise error("getaddrinfo returns an empty list")
    
    def to_bytes(s, encoding=None, errors='strict'):
        encoding = encoding or 'utf-8'
        if isinstance(s, bytes):
            if encoding != 'utf-8':
                return s.decode('utf-8', errors).encode(encoding, errors)
            else:
                return s
        else:
            return unicode(s).encode(encoding, errors)

    def to_string(s, encoding=None, errors='strict'):
        """Inverse of to_bytes"""
        if isinstance(s, bytes):
            return s.decode(encoding or 'utf-8', errors)
        else:
            return unicode(s)

    def native_str(s, encoding=None):
        if isinstance(s, unicode):
            return s.encode(encoding or 'utf-8')
        else:
            return s

    def force_native_str(s, encoding=None):
        if isinstance(s, unicode):
            return s.encode(encoding or 'utf-8')
        elif not isinstance(s, str):
            return str(s)
        else:
            return s

HTTPError = urllibr.HTTPError
URLError = urllibr.URLError
request_host = urllibr.request_host
parse_http_list = urllibr.parse_http_list

class SSLError(HTTPError):
    "Raised when SSL certificate fails in an HTTPS connection."
    pass

####################################################    URI & IRI SUFF
#
# The reserved URI characters (RFC 3986 - section 2.2)
#Default is charset is "iso-8859-1" (latin-1) from section 3.7.1
#http://www.ietf.org/rfc/rfc2616.txt
DEFAULT_CHARSET = 'iso-8859-1'
URI_GEN_DELIMS = frozenset(':/?#[]@')
URI_SUB_DELIMS = frozenset("!$&'()*+,;=")
URI_RESERVED_SET = URI_GEN_DELIMS.union(URI_SUB_DELIMS)
URI_RESERVED_CHARS = ''.join(URI_RESERVED_SET)
# The unreserved URI characters (RFC 3986 - section 2.3)
URI_UNRESERVED_SET = frozenset(ascii_letters + string.digits + '-._~')
URI_SAFE_CHARS = URI_RESERVED_CHARS + '%~'
HEADER_TOKEN_CHARS = frozenset("!#$%&'*+-.0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"
                               '^_`abcdefghijklmnopqrstuvwxyz|~')

escape = lambda s: quote(s, safe='~')
urlquote = lambda iri: quote(iri, safe=URI_RESERVED_CHARS)

def _gen_unquote(uri):
    unreserved_set = URI_UNRESERVED_SET
    for n, part in enumerate(force_native_str(uri).split('%')):
        if not n:
            yield part
        else:
            h = part[0:2]
            if len(h) == 2:
                c = chr(int(h, 16))
                if c in unreserved_set:
                    yield c + part[2:]
                else:
                    yield '%' + part
            else:
                yield '%' + part
                
def unquote_unreserved(uri):
    """Un-escape any percent-escape sequences in a URI that are unreserved
characters. This leaves all reserved, illegal and non-ASCII bytes encoded."""
    return ''.join(_gen_unquote(uri))

def requote_uri(uri):
    """Re-quote the given URI.

    This function passes the given URI through an unquote/quote cycle to
    ensure that it is fully and consistently quoted.
    """
    # Unquote only the unreserved characters
    # Then quote only illegal characters (do not quote reserved, unreserved,
    # or '%')
    return quote(unquote_unreserved(uri), safe=URI_SAFE_CHARS)

def iri_to_uri(iri, kwargs=None):
    '''Convert an Internationalized Resource Identifier (IRI) portion to a URI
portion that is suitable for inclusion in a URL.
This is the algorithm from section 3.1 of RFC 3987.
Returns an ASCII native string containing the encoded result.'''
    if iri is None:
        return iri
    iri = force_native_str(iri)
    if kwargs:
        iri = '%s?%s'%(iri,'&'.join(('%s=%s' % kv for kv in iteritems(kwargs))))
    return urlquote(unquote_unreserved(iri))

def host_and_port(host):
    host, port = splitport(host)
    return host, int(port) if port else None

def default_port(scheme):
    if scheme == "http":
        return '80'
    elif scheme == "https":
        return '443'
    
def host_and_port_default(scheme, host):
    host, port = splitport(host)
    if not port:
        port = default_port(scheme)
    return host, port

def get_hostport(scheme, full_host):
    host, port = host_and_port(full_host)
    if port is None:
        i = host.rfind(':')
        j = host.rfind(']')         # ipv6 addresses have [...]
        if i > j:
            try:
                port = int(host[i+1:])
            except ValueError:
                if host[i+1:] == "": # http://foo.com:/ == http://foo.com/
                    port = default_port(scheme)
                else:
                    raise httpclient.InvalidURL("nonnumeric port: '%s'"
                                                 % host[i+1:])
            host = host[:i]
        else:
            port = default_port(scheme)
        if host and host[0] == '[' and host[-1] == ']':
            host = host[1:-1]
    return host, port

def remove_double_slash(route):
    if '//' in route:
        route = re.sub('/+', '/', route)
    return route

def is_closed(sock):    #pragma nocover
    """Check if socket is connected."""
    if not sock:
        return False
    try:
        if not poll:
            if not select:
                return False
            try:
                return bool(select([sock], [], [], 0.0)[0])
            except socket.error:
                return True
        # This version is better on platforms that support it.
        p = poll()
        p.register(sock, POLLIN)
        for (fno, ev) in p.poll(0.0):
            if fno == sock.fileno():
                # Either data is buffered (bad), or the connection is dropped.
                return True
    except Exception:
        return True

####################################################    CONTENT TYPES
JSON_CONTENT_TYPES = ('application/json',
                      'application/javascript',
                      'text/json',
                      'text/x-json')

####################################################    REQUEST METHODS
ENCODE_URL_METHODS = frozenset(['DELETE', 'GET', 'HEAD', 'OPTIONS'])
ENCODE_BODY_METHODS = frozenset(['PATCH', 'POST', 'PUT', 'TRACE'])
REDIRECT_CODES = (301, 302, 303, 307)

def has_empty_content(status, method=None):
    '''204, 304 and 1xx codes have no content'''
    if status == httpclient.NO_CONTENT or\
            status == httpclient.NOT_MODIFIED or\
            100 <= status < 200 or\
            method == "HEAD":
        return True
    else:
        return False

def is_succesful(status):
    '''2xx status is succesful'''
    return status >= 200 and status < 300

####################################################    HTTP HEADERS
WEBSOCKET_VERSION = (8, 13)
HEADER_FIELDS = {'general': frozenset(('Cache-Control', 'Connection', 'Date',
                                       'Pragma', 'Trailer','Transfer-Encoding',
                                       'Upgrade', 'Sec-WebSocket-Extensions',
                                       'Sec-WebSocket-Protocol',
                                       'Via','Warning')),
                 # The request-header fields allow the client to pass additional
                 # information about the request, and about the client itself,
                 # to the server.
                 'request': frozenset(('Accept', 'Accept-Charset',
                                       'Accept-Encoding', 'Accept-Language',
                                       'Authorization',
                                       'Cookie', 'Expect', 'From',
                                       'Host', 'If-Match', 'If-Modified-Since',
                                       'If-None-Match', 'If-Range',
                                       'If-Unmodified-Since', 'Max-Forwards',
                                       'Proxy-Authorization', 'Range',
                                       'Referer',
                                       'Sec-WebSocket-Key',
                                       'Sec-WebSocket-Version',
                                       'TE',
                                       'User-Agent',
                                       'X-Requested-With')),
                 # The response-header fields allow the server to pass
                 # additional information about the response which cannot be
                 # placed in the Status- Line.
                 'response': frozenset(('Accept-Ranges',
                                        'Age',
                                        'ETag',
                                        'Location',
                                        'Proxy-Authenticate',
                                        'Retry-After',
                                        'Sec-WebSocket-Accept',
                                        'Server',
                                        'Set-Cookie',
                                        'Set-Cookie2',
                                        'Vary',
                                        'WWW-Authenticate',
                                        'X-Frame-Options')),
                 'entity': frozenset(('Allow', 'Content-Encoding',
                                      'Content-Language', 'Content-Length',
                                      'Content-Location', 'Content-MD5',
                                      'Content-Range', 'Content-Type',
                                      'Expires', 'Last-Modified'))}

CLIENT_HEADER_FIELDS = HEADER_FIELDS['general'].union(HEADER_FIELDS['entity'],
                                                      HEADER_FIELDS['request'])
SERVER_HEADER_FIELDS = HEADER_FIELDS['general'].union(HEADER_FIELDS['entity'],
                                                      HEADER_FIELDS['response'])
ALL_HEADER_FIELDS = CLIENT_HEADER_FIELDS.union(SERVER_HEADER_FIELDS)
ALL_HEADER_FIELDS_DICT = dict(((k.lower(),k) for k in ALL_HEADER_FIELDS))

TYPE_HEADER_FIELDS = {'client': CLIENT_HEADER_FIELDS,
                      'server': SERVER_HEADER_FIELDS,
                      'both': ALL_HEADER_FIELDS}

header_type = {0: 'client', 1: 'server', 2: 'both'}
header_type_to_int = dict(((v,k) for k,v in header_type.items()))

def capfirst(x):
    x = x.strip()
    if x:
        return x[0].upper() + x[1:].lower()
    else:
        return x
    
def capheader(name):
    return '-'.join((b for b in (capfirst(n) for n in name.split('-')) if b))

def header_field(name, HEADERS_SET=None, strict=False):
    name = name.lower()
    if name.startswith('x-'):
        return capheader(name)
    else:
        header = ALL_HEADER_FIELDS_DICT.get(name)
        if header and HEADERS_SET:
            return header if header in HEADERS_SET else None
        elif header:
            return header
        elif not strict:
            return capheader(name)

def quote_header_value(value, extra_chars='', allow_token=True):
    """Quote a header value if necessary.

:param value: the value to quote.
:param extra_chars: a list of extra characters to skip quoting.
:param allow_token: if this is enabled token values are returned
                    unchanged."""
    value = force_native_str(value)
    if allow_token:
        token_chars = HEADER_TOKEN_CHARS | set(extra_chars)
        if set(value).issubset(token_chars):
            return value
    return '"%s"' % value.replace('\\', '\\\\').replace('"', '\\"')

def unquote_header_value(value, is_filename=False):
    """Unquotes a header value. Reversal of :func:`quote_header_value`.
This does not use the real unquoting but what browsers are actually
using for quoting.
:param value: the header value to unquote."""
    if value and value[0] == value[-1] == '"':
        # this is not the real unquoting, but fixing this so that the
        # RFC is met will result in bugs with internet explorer and
        # probably some other browsers as well.  IE for example is
        # uploading files with "C:\foo\bar.txt" as filename
        value = value[1:-1]
        # if this is a filename and the starting characters look like
        # a UNC path, then just return the value without quotes.  Using the
        # replace sequence below on a UNC path has the effect of turning
        # the leading double slash into a single slash and then
        # _fix_ie_filename() doesn't work correctly.  See #458.
        if not is_filename or value[:2] != '\\\\':
            return value.replace('\\\\', '\\').replace('\\"', '"')
    return value

def parse_dict_header(value):
    """Parse lists of key, value pairs as described by RFC 2068 Section 2 and
    convert them into a python dict:

    >>> d = parse_dict_header('foo="is a fish", bar="as well"')
    >>> type(d) is dict
    True
    >>> sorted(d.items())
    [('bar', 'as well'), ('foo', 'is a fish')]

    If there is no value for a key it will be `None`:

    >>> parse_dict_header('key_without_value')
    {'key_without_value': None}

    To create a header from the :class:`dict` again, use the
    :func:`dump_header` function.

    :param value: a string with a dict header.
    :return: :class:`dict`
    """
    result = {}
    for item in parse_http_list(value):
        if '=' not in item:
            result[item] = None
            continue
        name, value = item.split('=', 1)
        if value[:1] == value[-1:] == '"':
            value = unquote_header_value(value[1:-1])
        result[name] = value
    return result


class Headers(object):
    '''Utility for managing HTTP headers for both clients and servers.
It has a dictionary like interface with few extra functions to facilitate
the insertion of multiple header values. Header fields are **case insensitive**,
therefore doing::

    >>> h = Headers()
    >>> h['Content-Length'] = '1050'
    
is equivalent to

    >>> h['content-length'] = '1050'

:param headers: optional iterable over header field/value pairs.
:param kind: optional headers type, one of ``server``, ``client`` or ``both``.
:param strict: if ``True`` only valid headers field will be included.

This :class:`Headers` container maintains an ordering as suggested by
http://www.w3.org/Protocols/rfc2616/rfc2616.html:

.. epigraph::

    The order in which header fields with differing field names are received is not
    significant. However, it is "good practice" to send general-header fields first,
    followed by request-header or response-header fields, and ending with
    the entity-header fields.
    
    -- rfc2616 section 4.2
    
The strict parameter is rarely used and it forces the omission on non-standard
header fields.
'''
    def __init__(self, headers=None, kind='server', strict=False):
        if isinstance(kind, int):
            kind = header_type.get(kind, 'both')
        else:
            kind = str(kind).lower()
        self.kind = kind
        self.strict = strict
        self.all_headers = TYPE_HEADER_FIELDS.get(self.kind)
        if not self.all_headers:
            self.kind = 'both'
            self.all_headers = TYPE_HEADER_FIELDS[self.kind]
        self._headers = {}
        if headers is not None:
            self.update(headers)

    def __repr__(self):
        return '%s %s' % (self.kind, self._headers.__repr__())

    def __str__(self):
        return '\r\n'.join(self._ordered())

    def __bytes__(self):
        return str(self).encode('latin-1')

    def __iter__(self):
        for k, values in iteritems(self._headers):
            yield k, ', '.join(values)

    def __len__(self):
        return len(self._headers)

    @property
    def kind_number(self):
        return header_type_to_int.get(self.kind)

    def update(self, iterable):
        """Extend the headers with a dictionary or an iterable yielding keys
and values."""
        for key, value in mapping_iterator(iterable):
            self[key] = value

    def __contains__(self, key):
        return header_field(key) in self._headers

    def __getitem__(self, key):
        return ', '.join(self._headers[header_field(key)])

    def __delitem__(self, key):
        self._headers.__delitem__(header_field(key))

    def __setitem__(self, key, value):
        key = header_field(key, self.all_headers, self.strict)
        if key and value is not None:
            if not isinstance(value, list):
                value = self.get_values(value)
            self._headers[key] = value

    def get(self, key, default=None):
        '''Get the value at header field ``key`` as a comma separated
string of values. For example::

    >>> from pulsar.utils.httpurl import Headers
    >>> h = Headers(kind='client')
    >>> h.add_header('accept-encoding', 'gzip')
    >>> h.add_header('accept-encoding', 'deflate')
    >>> h.get('accept-encoding')
    
results in::

    'gzip, deflate'
'''
        if key in self:
            return self.__getitem__(key)
        else:
            return default

    def get_all(self, key, default=None):
        '''Get the values at header field ``key`` as a list rather than a
string separated by comma (which is returned by the :meth:`get` method).
For example::

    >>> from pulsar.utils.httpurl import Headers
    >>> h = Headers(kind='client')
    >>> h.add_header('accept-encoding', 'gzip')
    >>> h.add_header('accept-encoding', 'deflate')
    >>> h.get_all('accept-encoding')
    
results in::

    ['gzip', 'deflate']
'''
        return self._headers.get(header_field(key), default)

    def has(self, field, value):
        '''Check if ``value`` is avialble in header ``field``.'''
        value = value.lower()
        for c in self.get_all(field, ()):
            if c.lower() == value:
                return True
        return False
    
    def pop(self, key, *args):
        return self._headers.pop(header_field(key), *args)

    def copy(self):
        return self.__class__(self, kind=self.kind)
    
    def clear(self):
        '''Same as :meth:`dict.clear`, it removes all headers.'''
        self._headers.clear()

    def getheaders(self, key):
        '''Required by cookielib. If the key is not available,
it returns an empty list.'''
        return self._headers.get(header_field(key), [])

    def headers(self):
        return list(self)

    def add_header(self, key, value, **params):
        '''Add *value* to *key* header. If the header is already available,
append the value to the list.'''
        key = header_field(key, self.all_headers, self.strict)
        if key and value:
            values = self._headers.get(key, [])
            lower_values = [v.lower() for v in values]
            for value in self.get_values(value):
                lower_value = value.lower() 
                if lower_value not in lower_values:
                    lower_values.append(lower_value)
                    values.append(value)
            self._headers[key] = values
                
    def remove_header(self, key, value=None):
        '''Remove the header at ``key``, If ``value`` is provided, it removes
only that value if found.'''
        key = header_field(key, self.all_headers, self.strict)
        if key:
            if value:
                value = value.lower()
                values = self._headers.get(key, [])
                removed = None
                for v in values:
                    if v.lower() == value:
                        removed = v
                        values.remove(v)
                self._headers[key] = values
                return removed
            else:
                return self._headers.pop(key, None)

    def flat(self, version, status):
    	'''Full headers bytes representation'''
    	vs = version + (status, self)
    	return ('HTTP/%s.%s %s\r\n%s' % vs).encode(DEFAULT_CHARSET)

    def _ordered(self):
        hf = HEADER_FIELDS
        order = (('general',[]), ('request',[]), ('response',[]), ('entity',[]))
        headers = self._headers
        for key in headers:
            for name, group in order:
                if key in hf[name]:
                    group.append(key)
                    break
            if key not in group:    # non-standard header
                group.append(key)
        for _, group in order:
            for k in group:
                yield "%s: %s" % (k, ', '.join(headers[k]))
        yield ''
        yield ''
        
    ##    INTERNALS
    def get_values(self, value):
        return [v for v in (v.strip() for v in value.split(',')) if v]


###############################################################################
##    HTTP PARSER
###############################################################################
METHOD_RE = re.compile("[A-Z0-9$-_.]{3,20}")
VERSION_RE = re.compile("HTTP/(\d+).(\d+)")
STATUS_RE = re.compile("(\d{3})\s*(\w*)")
HEADER_RE = re.compile("[\x00-\x1F\x7F()<>@,;:\[\]={} \t\\\\\"]")

# errors
BAD_FIRST_LINE = 0
INVALID_HEADER = 1
INVALID_CHUNK = 2

class InvalidRequestLine(Exception):
    """ error raised when first line is invalid """


class InvalidHeader(Exception):
    """ error raised on invalid header """


class InvalidChunkSize(Exception):
    """ error raised when we parse an invalid chunk size """


class HttpParser(object):
    '''A python HTTP parser.

Original code from https://github.com/benoitc/http-parser

2011 (c) Benoit Chesneau <benoitc@e-engura.org>

Permission is hereby granted, free of charge, to any person
obtaining a copy of this software and associated documentation
files (the "Software"), to deal in the Software without
restriction, including without limitation the rights to use,
copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.'''
    def __init__(self, kind=2, decompress=False):
        self.decompress = decompress
        # errors vars
        self.errno = None
        self.errstr = ""
        # protected variables
        self._buf = []
        self._version = None
        self._method = None
        self._status_code = None
        self._status = None
        self._reason = None
        self._url = None
        self._path = None
        self._query_string = None
        self._kind = kind
        self._fragment= None
        self._headers = OrderedDict()
        self._chunked = False
        self._body = []
        self._trailers = None
        self._partial_body = False
        self._clen = None
        self._clen_rest = None
        # private events
        self.__on_firstline = False
        self.__on_headers_complete = False
        self.__on_message_begin = False
        self.__on_message_complete = False
        self.__decompress_obj = None

    @property
    def kind(self):
        return self._kind

    def get_version(self):
        return self._version

    def get_method(self):
        return self._method

    def get_status_code(self):
        return self._status_code

    def get_url(self):
        return self._url

    def get_path(self):
        return self._path

    def get_query_string(self):
        return self._query_string

    def get_fragment(self):
        return self._fragment

    def get_headers(self):
        return self._headers

    def recv_body(self):
        """ return last chunk of the parsed body"""
        body = b''.join(self._body)
        self._body = []
        self._partial_body = False
        return body

    def is_headers_complete(self):
        """ return True if all headers have been parsed. """
        return self.__on_headers_complete

    def is_partial_body(self):
        """ return True if a chunk of body have been parsed """
        return self._partial_body

    def is_message_begin(self):
        """ return True if the parsing start """
        return self.__on_message_begin

    def is_message_complete(self):
        """ return True if the parsing is done (we get EOF) """
        return self.__on_message_complete

    def is_chunked(self):
        """ return True if Transfer-Encoding header value is chunked"""
        return self._chunked

    def execute(self, data, length):
        # end of body can be passed manually by putting a length of 0
        if length == 0:
            return self.close(length)
        data = bytes(data)
        # start to parse
        nb_parsed = 0
        while True:
            if not self.__on_firstline:
                idx = data.find(b'\r\n')
                if idx < 0:
                    self._buf.append(data)
                    return len(data)
                else:
                    self.__on_firstline = True
                    self._buf.append(data[:idx])
                    first_line = native_str(b''.join(self._buf),DEFAULT_CHARSET)
                    rest = data[idx+2:]
                    data = b''
                    if self._parse_firstline(first_line):
                        nb_parsed = nb_parsed + idx + 2
                        self._buf = [rest]
                    else:
                        return nb_parsed
            elif not self.__on_headers_complete:
                if data:
                    self._buf.append(data)
                    data = b''
                try:
                    to_parse = b''.join(self._buf)
                    ret = self._parse_headers(to_parse)
                    if ret is False:
                        if to_parse == b'\r\n':
                            self._buf = []
                            return self.close(length)
                        return length
                    nb_parsed = nb_parsed + (len(to_parse) - ret)
                except InvalidHeader as e:
                    self.errno = INVALID_HEADER
                    self.errstr = str(e)
                    return nb_parsed
            elif not self.__on_message_complete:
                self.__on_message_begin = True
                if data:
                    self._buf.append(data)
                    data = b''
                ret = self._parse_body()
                if ret is None:
                    return length
                elif ret < 0:
                    return ret
                elif ret == 0:
                    self.__on_message_complete = True
                    return length
                else:
                    nb_parsed = max(length, ret)
            else:
                return 0

    def close(self, length):
        self.__on_message_begin = True
        self.__on_message_complete = True
        if not self._buf and self.__on_firstline:
            self.__on_headers_complete = True
            return length
        else:
            return length+len(self._buf)
        
    def _parse_firstline(self, line):
        try:
            if self.kind == 2: # auto detect
                try:
                    self._parse_request_line(line)
                except InvalidRequestLine:
                    self._parse_response_line(line)
            elif self.kind == 1:
                self._parse_response_line(line)
            elif self.kind == 0:
                self._parse_request_line(line)
        except InvalidRequestLine as e:
            self.errno = BAD_FIRST_LINE
            self.errstr = str(e)
            return False
        return True

    def _parse_response_line(self, line):
        bits = line.split(None, 1)
        if len(bits) != 2:
            raise InvalidRequestLine(line)

        # version
        matchv = VERSION_RE.match(bits[0])
        if matchv is None:
            raise InvalidRequestLine("Invalid HTTP version: %s" % bits[0])
        self._version = (int(matchv.group(1)), int(matchv.group(2)))

        # status
        matchs = STATUS_RE.match(bits[1])
        if matchs is None:
            raise InvalidRequestLine("Invalid status %" % bits[1])

        self._status = bits[1]
        self._status_code = int(matchs.group(1))
        self._reason = matchs.group(2)

    def _parse_request_line(self, line):
        bits = line.split(None, 2)
        if len(bits) != 3:
            raise InvalidRequestLine(line)
        # Method
        if not METHOD_RE.match(bits[0]):
            raise InvalidRequestLine("invalid Method: %s" % bits[0])
        self._method = bits[0].upper()
        # URI
        self._url = bits[1]
        parts = urlsplit(bits[1])
        self._path = parts.path or ""
        self._query_string = parts.query or ""
        self._fragment = parts.fragment or ""
        # Version
        match = VERSION_RE.match(bits[2])
        if match is None:
            raise InvalidRequestLine("Invalid HTTP version: %s" % bits[2])
        self._version = (int(match.group(1)), int(match.group(2)))

    def _parse_headers(self, data):
        idx = data.find(b'\r\n\r\n')
        if idx < 0: # we don't have all headers
            return False
        chunk = native_str(data[:idx], DEFAULT_CHARSET)
        # Split lines on \r\n keeping the \r\n on each line
        lines = deque(('%s\r\n' % line for line in chunk.split('\r\n')))
        # Parse headers into key/value pairs paying attention
        # to continuation lines.
        while len(lines):
            # Parse initial header name : value pair.
            curr = lines.popleft()
            if curr.find(":") < 0:
                continue
            name, value = curr.split(":", 1)
            name = name.rstrip(" \t").upper()
            if HEADER_RE.search(name):
                raise InvalidHeader("invalid header name %s" % name)
            name, value = name.strip().lower(), [value.lstrip()]
            # Consume value continuation lines
            while len(lines) and lines[0].startswith((" ", "\t")):
                value.append(lines.popleft())
            value = ''.join(value).rstrip()
            # multiple headers
            if name in self._headers:
                value = "%s, %s" % (self._headers[name], value)
            # store new header value
            self._headers[name] = value
        # detect now if body is sent by chunks.
        clen = self._headers.get('content-length')
        te = self._headers.get('transfer-encoding', '').lower()
        self._chunked = (te == 'chunked')
        if clen and not self._chunked:
            try:
                clen = int(clen)
            except ValueError:
                clen = None
            else:
                if clen < 0:  # ignore nonsensical negative lengths
                    clen = None
        else:
            clen = None
        status = self._status_code
        if status and (status == httpclient.NO_CONTENT or
            status == httpclient.NOT_MODIFIED or
            100 <= status < 200 or      # 1xx codes
            self._method == "HEAD"):
            clen = 0
        self._clen_rest = self._clen = clen
        # detect encoding and set decompress object
        if self.decompress:
            encoding = self._headers.get('content-encoding')
            if encoding == "gzip":
                self.__decompress_obj = zlib.decompressobj(16+zlib.MAX_WBITS)
            elif encoding == "deflate":
                self.__decompress_obj = zlib.decompressobj()
        rest = data[idx+4:]
        self._buf = [rest]
        self.__on_headers_complete = True
        return len(rest)

    def _parse_body(self):
        data = b''.join(self._buf)
        if not self._chunked:
            if self._clen_rest is not None:
                self._clen_rest -= len(data)
            # maybe decompress
            if self.__decompress_obj is not None:
                data = self.__decompress_obj.decompress(data)
            self._partial_body = True
            if data:
                self._body.append(data)
            self._buf = []
            if self._clen_rest is None or self._clen_rest <= 0:
                self.__on_message_complete = True
        else:
            try:
                size, rest = self._parse_chunk_size(data)
            except InvalidChunkSize as e:
                self.errno = INVALID_CHUNK
                self.errstr = "invalid chunk size [%s]" % str(e)
                return -1
            if size == 0:
                return size
            if size is None or len(rest) < size:
                return None
            body_part, rest = rest[:size], rest[size:]
            if len(rest) < 2:
                self.errno = INVALID_CHUNK
                self.errstr = "chunk missing terminator [%s]" % data
                return -1
            # maybe decompress
            if self.__decompress_obj is not None:
                body_part = self.__decompress_obj.decompress(body_part)
            self._partial_body = True
            self._body.append(body_part)
            self._buf = [rest[2:]]
            return len(rest)

    def _parse_chunk_size(self, data):
        idx = data.find(b'\r\n')
        if idx < 0:
            return None, None
        line, rest_chunk = data[:idx], data[idx+2:]
        chunk_size = line.split(b';', 1)[0].strip()
        try:
            chunk_size = int(chunk_size, 16)
        except ValueError:
            raise InvalidChunkSize(chunk_size)
        if chunk_size == 0:
            self._parse_trailers(rest_chunk)
            return 0, None
        return chunk_size, rest_chunk

    def _parse_trailers(self, data):
        idx = data.find(b'\r\n\r\n')
        if data[:2] == b'\r\n':
            self._trailers = self._parse_headers(data[:idx])

if not hasextensions:
    setDefaultHttpParser(HttpParser)
    
################################################################################
##    HTTP CLIENT
################################################################################

class WWWAuthenticate(object):
    _require_quoting = frozenset(['domain', 'nonce', 'opaque', 'realm'])
    
    def __init__(self, type, **options):
        self.type = type
        self.options = options
    
    @classmethod
    def basic(cls, realm='authentication required'):
        """Clear the auth info and enable basic auth."""
        return cls('basic', realm=realm)
    
    @classmethod    
    def digest(cls, realm, nonce, qop=('auth',), opaque=None,
               algorithm=None, stale=None):
        options = {}
        if stale:
            options['stale'] = 'TRUE'
        if opaque is not None:
            options['opaque'] = opaque
        if algorithm is not None:
            options['algorithm'] = algorithm
        return cls('digest', realm=realm, nonce=nonce,
                   qop=', '.join(qop), **options)
    
    def __str__(self):
        """Convert the stored values into a WWW-Authenticate header."""
        return '%s %s' % (self.type.title(), ', '.join((
            '%s=%s' % (key, quote_header_value(value,
                                allow_token=key not in self._require_quoting))
            for key, value in iteritems(self.options)
        )))
    __repr__ = __str__
        
###############################################    UTILITIES, ENCODERS, PARSERS
def get_environ_proxies():
    """Return a dict of environment proxies. From requests_."""

    proxy_keys = [
        'all',
        'http',
        'https',
        'ftp',
        'socks',
        'no'
    ]

    get_proxy = lambda k: os.environ.get(k) or os.environ.get(k.upper())
    proxies = [(key, get_proxy(key + '_proxy')) for key in proxy_keys]
    return dict([(key, val) for (key, val) in proxies if val])

def addslash(url):
    '''Add a slash at the beginning of *url*.'''
    if not url.startswith('/'):
        url = '/%s' % url
    return url

def appendslash(url):
    '''Append a slash to *url* if it does not have one.'''
    if not url.endswith('/'):
        url = '%s/' % url
    return url

def choose_boundary():
    """Our embarassingly-simple replacement for mimetools.choose_boundary."""
    return uuid4().hex

def get_content_type(filename):
    return mimetypes.guess_type(filename)[0] or 'application/octet-stream'

def encode_multipart_formdata(fields, boundary=None, charset=None):
    """
    Encode a dictionary of ``fields`` using the multipart/form-data mime format.

    :param fields:
        Dictionary of fields or list of (key, value) field tuples.  The key is
        treated as the field name, and the value as the body of the form-data
        bytes. If the value is a tuple of two elements, then the first element
        is treated as the filename of the form-data section.

        Field names and filenames must be unicode.

    :param boundary:
        If not specified, then a random boundary will be generated using
        :func:`mimetools.choose_boundary`.
    """
    charset = charset or 'utf-8'
    body = BytesIO()
    if boundary is None:
        boundary = choose_boundary()
    for fieldname, value in mapping_iterator(fields):
        body.write(('--%s\r\n' % boundary).encode(charset))
        if isinstance(value, tuple):
            filename, data = value
            body.write(('Content-Disposition: form-data; name="%s"; '
                        'filename="%s"\r\n' % (fieldname, filename))\
                       .encode(charset))
            body.write(('Content-Type: %s\r\n\r\n' %
                       (get_content_type(filename))).encode(charset))
        else:
            data = value
            body.write(('Content-Disposition: form-data; name="%s"\r\n'
                        % (fieldname)).encode(charset))
            body.write(b'Content-Type: text/plain\r\n\r\n')
        data = body.write(to_bytes(data))
        body.write(b'\r\n')
    body.write(('--%s--\r\n' % (boundary)).encode(charset))
    content_type = 'multipart/form-data; boundary=%s' % boundary
    return body.getvalue(), content_type

def hexmd5(x):
    return md5(to_bytes(x)).hexdigest()

def hexsha1(x):
    return sha1(to_bytes(x)).hexdigest()

def http_date(epoch_seconds=None):
    """
    Formats the time to match the RFC1123 date format as specified by HTTP
    RFC2616 section 3.3.1.

    Accepts a floating point number expressed in seconds since the epoch, in
    UTC - such as that outputted by time.time(). If set to None, defaults to
    the current time.

    Outputs a string in the format 'Wdy, DD Mon YYYY HH:MM:SS GMT'.
    """
    return formatdate(epoch_seconds, usegmt=True)

#################################################################### COOKIE
def create_cookie(name, value, **kwargs):
    """Make a cookie from underspecified parameters.

    By default, the pair of `name` and `value` will be set for the domain ''
    and sent on every request (this is sometimes called a "supercookie").
    """
    result = dict(
        version=0,
        name=name,
        value=value,
        port=None,
        domain='',
        path='/',
        secure=False,
        expires=None,
        discard=True,
        comment=None,
        comment_url=None,
        rest={'HttpOnly': None},
        rfc2109=False,)
    badargs = set(kwargs) - set(result)
    if badargs:
        err = 'create_cookie() got unexpected keyword arguments: %s'
        raise TypeError(err % list(badargs))
    result.update(kwargs)
    result['port_specified'] = bool(result['port'])
    result['domain_specified'] = bool(result['domain'])
    result['domain_initial_dot'] = result['domain'].startswith('.')
    result['path_specified'] = bool(result['path'])
    return Cookie(**result)

def cookiejar_from_dict(cookie_dict, cookiejar=None):
    """Returns a CookieJar from a key/value dictionary.

    :param cookie_dict: Dict of key/values to insert into CookieJar.
    """
    if cookiejar is None:
        cookiejar = CookieJar()
    if cookie_dict is not None:
        for name in cookie_dict:
            cookiejar.set_cookie(create_cookie(name, cookie_dict[name]))
    return cookiejar

def parse_cookie(cookie):
    '''Parse an `HTTP cookie`_ string and return a dictionary of
cookie name/values.'''
    if cookie == '':
        return {}
    if not isinstance(cookie, BaseCookie):
        try:
            c = SimpleCookie()
            c.load(cookie)
        except CookieError:
            # Invalid cookie
            return {}
    else:
        c = cookie
    cookiedict = {}
    for key in c.keys():
        cookiedict[key] = c.get(key).value
    return cookiedict


cc_delim_re = re.compile(r'\s*,\s*')


class accept_content_type(object):

    def __init__(self, values=None):
        self._all = {}
        self.update(values)

    def update(self, values):
        if values:
            all = self._all
            accept_headers = cc_delim_re.split(values)
            for h in accept_headers:
                v = h.split('/')
                if len(v) == 2:
                    a, b = v
                    if a in all:
                        all[a].append(b)
                    else:
                        all[a] = [b]

    def __contains__(self, content_type):
        all = self._all
        if not all:
            # If no Accept header field is present, then it is assumed that the
            # client accepts all media types.
            return True
        a, b = content_type.split('/')
        if a in all:
            all = all[a]
            if '*' in all:
                return True
            else:
                return b in all
        elif '*' in all:
            return True
        else:
            return False


def patch_vary_headers(response, newheaders):
    """\
Adds (or updates) the "Vary" header in the given HttpResponse object.
newheaders is a list of header names that should be in "Vary". Existing
headers in "Vary" aren't removed.

For information on the Vary header, see:

    http://www.w3.org/Protocols/rfc2616/rfc2616-sec14.html#sec14.44
    """
    # Note that we need to keep the original order intact, because cache
    # implementations may rely on the order of the Vary contents in, say,
    # computing an MD5 hash.
    if 'Vary' in response:
        vary_headers = cc_delim_re.split(response['Vary'])
    else:
        vary_headers = []
    # Use .lower() here so we treat headers as case-insensitive.
    existing_headers = set([header.lower() for header in vary_headers])
    additional_headers = [newheader for newheader in newheaders
                          if newheader.lower() not in existing_headers]
    response['Vary'] = ', '.join(vary_headers + additional_headers)


def has_vary_header(response, header_query):
    """
    Checks to see if the response has a given header name in its Vary header.
    """
    if not response.has_header('Vary'):
        return False
    vary_headers = cc_delim_re.split(response['Vary'])
    existing_headers = set([header.lower() for header in vary_headers])
    return header_query.lower() in existing_headers


class CacheControl(object):
    '''
    http://www.mnot.net/cache_docs/

.. attribute:: maxage

    Specifies the maximum amount of time that a representation will be
    considered fresh.
    '''
    def __init__(self, maxage=None, private=False,
                 must_revalidate=False, proxy_revalidate=False,
                 nostore=False):
        self.maxage = maxage
        self.private = private
        self.must_revalidate = must_revalidate
        self.proxy_revalidate = proxy_revalidate
        self.nostore = nostore

    def __call__(self, headers):
        if self.nostore:
            headers['cache-control'] = 'no-store'
        elif self.maxage:
            headers['cache-control'] = 'max-age=%s' % self.maxage
            if self.private:
                headers.add_header('cache-control', 'private')
            else:
                headers.add_header('cache-control', 'public')
            if self.must_revalidate:
                headers.add_header('cache-control', 'must-revalidate')
            elif self.proxy_revalidate:
                headers.add_header('cache-control', 'proxy-revalidate')
        else:
            headers['cache-control'] = 'no-cache'




import math
import sys

if sys.version_info >= (3, 0):
    from itertools import zip_longest
else:   #pragma    nocover
    from itertools import izip_longest as zip_longest, izip as zip

__all__ = ['grouper', 'num2eng', 'nice_number']

 
# Tokens from 1000 and up
_PRONOUNCE = ['', 'thousand', 'million', 'billion', 'trillion']
 
# Tokens up to 90
_SMALL = {
    '1' : 'one',
    '2' : 'two',
    '3' : 'three',
    '4' : 'four',
    '5' : 'five',
    '6' : 'six',
    '7' : 'seven',
    '8' : 'eight',
    '9' : 'nine',
    '10' : 'ten',
    '11' : 'eleven',
    '12' : 'twelve',
    '13' : 'thirteen',
    '14' : 'fourteen',
    '15' : 'fifteen',
    '16' : 'sixteen',
    '17' : 'seventeen',
    '18' : 'eighteen',
    '19' : 'nineteen',
    '20' : 'twenty',
    '30' : 'thirty',
    '40' : 'forty',
    '50' : 'fifty',
    '60' : 'sixty',
    '70' : 'seventy',
    '80' : 'eighty',
    '90' : 'ninety'
}

def grouper(n, iterable, padvalue=None):
    "grouper(3, 'abcdefg', 'x') --> ('a','b','c'), ('d','e','f'), ('g','x','x')"
    return zip_longest(*[iter(iterable)]*n, fillvalue=padvalue)
 
def num2eng(num):
    '''English representation of a number up to a trillion'''
    num = str(int(num)) # Convert to string, throw if bad number
    if (len(num) / 3 >= len(_PRONOUNCE)): # Sanity check
        return num
    elif num == '0': # Zero is a special case
        return 'zero'
    pron = [] # Result accumulator
    first = True
    for pr, bits in zip(_PRONOUNCE, grouper(3, reversed(num), '')):
        num = ''.join(reversed(bits))
        n = int(num)
        bits = []
        if n > 99: # Got hundred
            bits.append('%s hundred' % _SMALL[num[0]])
            num = num[1:]
            n = int(num)
            num = str(n)
        if (n > 20) and (n != (n // 10 * 10)):
            bits.append('%s %s' % (_SMALL[num[0] + '0'], _SMALL[num[1]]))
        elif n:
            bits.append(_SMALL[num])
        if len(bits) == 2 and first:
            first = False
            p = ' and '.join(bits)
        else:
            p = ' '.join(bits)
        if p and pr:
            p = '%s %s' % (p, pr)
        pron.append(p)
    return ', '.join(reversed(pron))

def nice_number(number, name=None, plural=None):
    if not name:
        return num2eng(number)
    elif number == 1:
        return 'one %s' % name
    else:
        if not plural:
            plural = '%ss' % name
        return '%s %s' % (num2eng(number), plural)
    


'''Tests the websocket middleware in pulsar.apps.ws.'''
from pulsar import send
from pulsar.apps.ws import WebSocket
from pulsar.apps.http import HttpClient
from pulsar.apps.test import unittest, dont_run_with_thread

from .manage import server


class TestWebSocketThread(unittest.TestCase):
    app = None
    concurrency = 'thread'
    
    @classmethod
    def setUpClass(cls):
        s = server(bind='127.0.0.1:0', name=cls.__name__,
                   concurrency=cls.concurrency)
        cls.app = yield send('arbiter', 'run', s)
        cls.uri = 'http://{0}:{1}'.format(*cls.app.address)
        cls.ws_uri = 'ws://{0}:{1}/data'.format(*cls.app.address)
        cls.ws_echo = 'ws://{0}:{1}/echo'.format(*cls.app.address)
        
    @classmethod
    def tearDownClass(cls):
        if cls.app is not None:
            yield send('arbiter', 'kill_actor', cls.app.name)
    
    def testHyBiKey(self):
        w = WebSocket('/', None)
        v = w.challenge_response('dGhlIHNhbXBsZSBub25jZQ==')
        self.assertEqual(v, "s3pPLMBiTxaQ9kYGzzhZRbK+xOo=")
        
    def testBadRequests(self):
        c = HttpClient()
        response = yield c.post(self.ws_uri).on_finished
        self.assertEqual(response.status_code, 405)
        #
        response = yield c.get(self.ws_uri,
                        headers=[('Sec-Websocket-Key', '')]).on_finished
        self.assertEqual(response.status_code, 400)
        #
        response = yield c.get(self.ws_uri,
                        headers=[('Sec-Websocket-Key', 'bla')]).on_finished
        self.assertEqual(response.status_code, 400)
        #
        response = yield c.get(self.ws_uri,
                        headers=[('Sec-Websocket-version', 'xxx')]).on_finished
        self.assertEqual(response.status_code, 400)
    
    def testUpgrade(self):
        c = HttpClient()
        response = c.get(self.ws_echo)
        yield response.on_finished
        ws = response.on_finished.result
        response = ws.handshake 
        self.assertEqual(response.status_code, 101)
        self.assertEqual(response.headers['upgrade'], 'websocket')
        # Send a message to the websocket
        ws.write('Hi there!')
        

@dont_run_with_thread
class TestWebSocketProcess(TestWebSocketThread):
    concurrency = 'process'
    


from stdnet import getdb

from pulsar.apps import pubsub
from pulsar.utils.log import local_property


class PubSubBackend(pubsub.PubSubBackend):
    '''Implements :class:`PubSub` using a redis backend.'''
    @local_property
    def redis(self):
        redis = getdb(self.connection_string, timeout=0).client.pubsub()
        redis.bind_event('on_message', self.on_message)
        return redis
    
    def publish(self, channel, message):
        return self.redis.publish(channel, message)
       
    def subscribe(self, *channels):
        return self.redis.subscribe(*channels)
    
    def unsubscribe(self, *channels):
        return self.redis.unsubscribe(*channels)
    
    def on_message(self, channel_message):
        self.broadcast(*channel_message)


'''Utilities for HTML and text manipulation.

.. autofunction:: escape

.. autofunction:: mark_safe

.. autofunction:: slugify

.. autofunction:: capfirst

.. autofunction:: nicename

.. autofunction:: tag_attributes
'''
import re
import json
from unicodedata import normalize
from collections import namedtuple

from .pep import ispy3k, native_str, to_string, iteritems, is_string, string_type

NOTHING = ('', b'', None)
INLINE_TAGS = set(('input', 'meta', 'hr'))
DEFAULT_HTML_ATTRIBUTES = ('id', 'title')
HTML_ATTRIBUTES = {}
HTML_CHILDREN_TAG = {}

e = lambda *t: DEFAULT_HTML_ATTRIBUTES + t
input_attr = lambda *t: e('type', 'autocomplete', 'autofocus', 'disabled', 'form',
                          'formnovalidate', 'list', 'max', 'maxlength', 'min',
                          'multiple', 'name', 'pattern', 'placeholder',
                          'required', 'size', 'step', 'value', *t)

## HTML TAG ATTRIBUTES
################################################################################
HTML_ATTRIBUTES['a'] = e('href', 'name', 'target')
HTML_ATTRIBUTES['form'] = e('accept-charset', 'action', 'autocomplete',
                            'enctype', 'method', 'name', 'novalidate', 'target')
HTML_ATTRIBUTES['input'] = input_attr()
HTML_ATTRIBUTES['input[type="checkbox"]'] = input_attr('checked')
HTML_ATTRIBUTES['input[type="file"]'] = input_attr('accept')
HTML_ATTRIBUTES['input[type="image"]'] = input_attr('alt', 'formaction',
                                    'formenctype', 'formmethod', 'formtarget',
                                    'height', 'src', 'width')
HTML_ATTRIBUTES['input[type="radio"]'] = input_attr('checked')
HTML_ATTRIBUTES['input[type="submit"]'] = input_attr('formaction',
                                    'formenctype', 'formmethod', 'formtarget')
HTML_ATTRIBUTES['meta'] = e('name', 'charset', 'content')
HTML_ATTRIBUTES['option'] = e('disabled', 'label', 'selected', 'value')
HTML_ATTRIBUTES['script'] = e('async', 'charset', 'defer', 'src', 'type')
HTML_ATTRIBUTES['select'] = e('autofocus', 'disabled', 'form', 'multiple',
                              'name', 'required', 'size')
HTML_ATTRIBUTES['textarea'] = e('autofocus', 'cols', 'disabled', 'maxlength',
                                'name', 'placeholder', 'readonly', 'required',
                                'rows', 'wrap')
HTML_ATTRIBUTES['th'] = e('colspan', 'headers', 'rowspan', 'scope')

## DEFAULT HTML TAG CHILDREN
################################################################################
HTML_CHILDREN_TAG['ul'] = 'li'
HTML_CHILDREN_TAG['ol'] = 'li'
HTML_CHILDREN_TAG['select'] = 'option'

csslink = namedtuple('cssentry', 'link condition')

def tag_attributes(tag, type=None):
    '''Return a tuple of valid attributes for the HTML ``tag`` and optional
``type``. If the ``tag`` is not found in the global ``HTML_ATTRIBUTES``
dictionary, the ``DEFAULT_HTML_ATTRIBUTES`` set is returned.'''
    if type:
        ntag = '%s[type="%s"]' % (tag, type)
        if ntag in HTML_ATTRIBUTES:
            return HTML_ATTRIBUTES[ntag]
    return HTML_ATTRIBUTES.get(tag, DEFAULT_HTML_ATTRIBUTES)
    
def child_tag(tag):
    return HTML_CHILDREN_TAG.get(tag)

def slugify(value, rtx='_'):
    '''Normalizes string, removes non-alpha characters,
and converts spaces to ``rtx`` character (hyphens or underscore).'''
    value = normalize('NFKD', to_string(value)).encode('ascii', 'ignore')
    value = to_string(re.sub('[^\w\s-]', rtx, value.decode()).strip())
    return re.sub('[-\s]+', rtx, value)

def mark_safe(v):
    '''Mar a string ``v`` as safe. A safe string won't be escaped by the
:func:`escape` function.'''
    return SafeString(v)

def escape(html, force=False):
    """Returns the given HTML with ampersands,
quotes and angle brackets encoded."""
    if hasattr(html, '__html__') and not force:
        return html
    if html in NOTHING:
        return ''
    else:
        return to_string(html).replace('&', '&amp;').replace('<', '&lt;')\
            .replace('>', '&gt;').replace('"', '&quot;').replace("'", '&#39;')
            
def attr_iter(attrs):
    for k, v in iteritems(attrs):
        if v is not None:
            yield " %s='%s'" % (k, escape(v, force=True))

def dump_data_value(v):
    if not is_string(v):
        if isinstance(v, bytes):
            v = v.decode('utf-8')
        else:
            v = json.dumps(v)
    return mark_safe(v)

def lazy_string(f):
    def _(*args, **kwargs):
        return _lazy(f, args, kwargs)
    return _
    
def capfirst(x):
    '''Capitalise the first letter of ``x``.'''
    x = to_string(x).strip()
    if x:
        return x[0].upper() + x[1:]
    else:
        return x
    
def nicename(name):
    '''Make ``name`` a more user friendly string.'''
    name = to_string(name)
    return capfirst(' '.join(name.replace('-',' ').replace('_',' ').split()))


class SafeString(string_type):
    __html__ = True
    
class _lazy:

    def __init__(self, f, args, kwargs):
        self._value = None
        self._f = f
        self.args = args
        self.kwargs = kwargs
    
    def __str__(self):
        if self._value is None:
            self._value = native_str(self._f(*self.args, **self.kwargs) or '')
        return self._value
    __repr__ = __str__

if ispy3k:

    class UnicodeMixin(object):
        
        def __unicode__(self):
            return '%s object' % self.__class__.__name__

        def __str__(self):
            return self.__unicode__()

        def __repr__(self):
            return '%s: %s' % (self.__class__.__name__, self)

else: # Python 2    # pragma nocover

    class UnicodeMixin(object):

        def __unicode__(self):
            return unicode('%s object' % self.__class__.__name__)

        def __str__(self):
            return self.__unicode__().encode()

        def __repr__(self):
            return '%s: %s' % (self.__class__.__name__, self)


'''An asynchronous multi-process `HTTP proxy server`_ with *headers middleware*
to manipulate the original request headers. If the header middleware is
an empty list, the proxy passes requests and responses unmodified.
This is an implementation for a forward-proxy which can be used
to retrieve from any type of source from the Internet.

To run the server::

    python manage.py
    
An header middleware is a callable which receives the wsgi *environ* and
the list of request *headers*. By default the example uses:

.. autofunction:: x_forwarded_for

To run with different headers middleware create a new script and do::

    from proxyserver.manage import server
    
    if __name__ == '__main__':
        server(headers_middleware=[...]).start()
        
Implemenation
===========================

.. autoclass:: ProxyServerWsgiHandler
   :members:
   :member-order:
   
   
.. _`HTTP proxy server`: http://en.wikipedia.org/wiki/Proxy_server
'''
import io
import sys
from collections import deque
from functools import partial

try:
    import pulsar
except ImportError:
    sys.path.append('../../')

from pulsar import Deferred, HttpException
from pulsar.apps import wsgi, http
from pulsar.utils.httpurl import Headers
from pulsar.utils.log import LocalMixin, local_property

ENVIRON_HEADERS = ('content-type', 'content-length')
USER_AGENT = 'Pulsar-Proxy-Server'


def x_forwarded_for(environ, headers):
    '''Add *x-forwarded-for* header'''
    headers.add_header('x-forwarded-for', environ['REMOTE_ADDR'])
    
class user_agent:
    '''Override user-agent header'''
    def __init__(self, agent):
        self.agent = agent
        
    def __call__(self, environ, headers):
        headers['user-agent'] = self.agent

    
class ProxyServerWsgiHandler(LocalMixin):
    '''WSGI middleware for an asynchronous proxy server. To perform
processing on headers you can pass a list of ``headers_middleware``.
An headers middleware is a callable which accepts two parameters, the wsgi
*environ* dictionary and the *headers* container.'''
    def __init__(self, headers_middleware=None):
        self.headers_middleware = headers_middleware or []
    
    @local_property
    def http_client(self):
        return http.HttpClient(decompress=False, store_cookies=False)
        
    def __call__(self, environ, start_response):
        # The WSGI thing
        uri = environ['RAW_URI']
        if not uri or uri.startswith('/'):  # No proper uri, raise 404
            raise HttpException(status=404)
        request_headers = self.request_headers(environ)
        method = environ['REQUEST_METHOD']
        stream = environ.get('wsgi.input') or io.BytesIO()
        wsgi_response = ProxyResponse(start_response)
        response = self.http_client.request(method, uri,
                                            data=stream.getvalue(),
                                            headers=request_headers,
                                            version=environ['SERVER_PROTOCOL'])
        response.on_finished.add_errback(partial(wsgi_response.error, uri))
        response.bind_event('data_processed', wsgi_response)
        return wsgi_response
    
    def request_headers(self, environ):
        '''Fill request headers from the environ dictionary and
modify the headers via the list of :attr:`headers_middleware`.
The returned headers will be sent to the target uri.'''
        headers = Headers(kind='client')
        for k in environ:
            if k.startswith('HTTP_'):
                head = k[5:].replace('_','-')
                headers[head] = environ[k]
        for head in ENVIRON_HEADERS:
            k = head.replace('-','_').upper()
            v = environ.get(k)
            if v:
                headers[head] = v
        for middleware in self.headers_middleware:
            middleware(environ, headers)
        return headers


class Queue:
    
    def __init__(self):
        self._futures_in = deque()
        self._futures_out = deque()
        
    def size(self):
        return len(self._futures_out)
    
    def put(self, result):
        '''Put a result in the queue.'''
        future = self._futures_in.popleft()
        future.callback(result)
        
    def get(self):
        '''get a future from the queue.'''
        return self._futures_out.popleft()
        
    def add(self):
        '''Add a future to the queue'''
        future = Deferred()
        self._futures_in.append(future)
        self._futures_out.append(future)
        
    
class ProxyResponse(object):
    '''Asynchronous wsgi response.'''
    def __init__(self, start_response):
        self.start_response = start_response
        self.headers = None
        self.futures = Queue()
        self.futures.add()
        
    def __iter__(self):
        while self.futures.size():
            yield self.futures.get()
            
    def __call__(self, response):
        try:
            result = self._body(response)
        except Exception:
            result = sys.exc_info()
        if result is not None:
            if not response.parser.is_message_complete():
                self.futures.add()
            self.futures.put(result)
        
    def error(self, uri, failure):
        '''Handle a failure.'''
        failure.log()
        msg = 'Oops! Could not find %s' % uri
        html = wsgi.HtmlDocument(title=msg)
        html.body.append('<h1>%s</h1>' % msg)
        data = html.render()
        response = wsgi.WsgiResponse(504, data, content_type='text/html',
                                environ={}, start_response=self.start_response)
        response.start()
        self.futures.put(response.content[0])
        
    def _body(self, response):
        if response.parser.is_headers_complete():
            if self.headers is None:
                headers = self.remove_hop_headers(response.headers)
                self.headers = Headers(headers, kind='server')
                # start the response
                self.start_response(response.status, list(self.headers))
            return response.recv_body()
    
    def remove_hop_headers(self, headers):
        for header, value in headers:
            if header.lower() not in wsgi.HOP_HEADERS:
                yield header, value
                

def server(name='proxy-server', headers_middleware=None, **kwargs):
    if headers_middleware is None:
        headers_middleware = [user_agent(USER_AGENT), x_forwarded_for]
    wsgi_proxy = ProxyServerWsgiHandler(headers_middleware)
    return wsgi.WSGIServer(wsgi_proxy, name=name, **kwargs)


if __name__ == '__main__':
    server().start()



'''
The dining philosophers problem is an example problem often used in concurrent
algorithm design to illustrate synchronization issues and techniques
for resolving them.

The problem
===================

Five silent philosophers sit at a round table with each a bowl of spaghetti.
A fork is placed between each pair of adjacent philosophers.


         P     P
         O  f  O
        f       f
     P O         O P
         f     f 
            O
            P
           
Each philosopher ``P`` must alternately think and eat from his bowl ``O``.
Eating is not limited by the amount of spaghetti left: assume an infinite
supply.

However, a philosopher can only eat while holding both the fork ``f`` to
the left and the fork to the right.
Each philosopher can pick up an adjacent fork, when available, and put it down,
when holding it. These are separate actions: forks must be picked up and put
down one by one.

This implementation will just work. No starvation or dead-lock.

There are two parameters:

* Average eating period, the higher the more time is spend eating.
* Average waiting period, the higher the more frequent philosophers
    get a chance to eat.
    
To run the example, simply type::

    pulsar manage.py
'''
import random
import time
try:
    import pulsar
except ImportError:
    import sys
    sys.path.append('../../')
    import pulsar
from pulsar import command
    
################################################################################
##    EXTRA COMMAND LINE PARAMETERS
class Eating_Period(pulsar.Setting):
    flags = ["--eating-period"]
    validator = pulsar.validate_pos_float
    default = 2
    desc = """The average period of eating for a philosopher."""
    
    
class Waiting_Period(pulsar.Setting):
    flags = ["--waiting-period"]
    validator = pulsar.validate_pos_float
    default = 2
    desc = """The average period of waiting for a missing fork."""
    
################################################################################
##    PULSAR COMMANDS FOR DINING PHILOSOPHERS
@command(ack=False)
def putdown_fork(request, fork):
    self = request.actor.app
    try:
        self.not_available_forks.remove(fork)
    except KeyError:
        self.logger.error('Putting down a fork which was already available')    

@command()
def pickup_fork(request, fork_right):
    self = request.actor.app
    num_philosophers = self.cfg.workers
    fork_left = fork_right - 1
    if fork_left == 0:
        fork_left = num_philosophers 
    for fork in (fork_right, fork_left):
        if fork not in self.not_available_forks:
            # Fork is available, send it to the philosopher
            self.not_available_forks.add(fork)
            return fork


################################################################################
##    DINING PHILOSOPHERS APP
class DiningPhilosophers(pulsar.Application):
    description = 'Dining philosophers sit at a table around a bowl of '\
                  'spaghetti and waits for available forks.'
    cfg = pulsar.Config(workers=5)
    
    def monitor_start(self, monitor):
        self.not_available_forks = set()
        
    def worker_start(self, philosopher):
        self.take_action(philosopher)
    
    def take_action(self, philosopher):
        params = philosopher.params
        eaten = params.eaten or 0
        forks = params.forks
        started_waiting = params.started_waiting or 0
        pick_up_fork = True
        if forks:
            max_eat_period = 2*self.cfg.eating_period
            # Two forks. Eat!
            if len(forks) == 2:
                params.thinking = 0
                eaten += 1
                philosopher.logger.info("%s eating... So far %s times",
                                        philosopher.name, eaten)
                try:
                    time.sleep(max_eat_period*random.random())
                except IOError:
                    pass
                params.eaten = eaten
                pick_up_fork = False
            # One fork only! release fork or try to pick up one
            elif len(forks) == 1:
                waiting_period = 2*self.cfg.waiting_period*random.random()
                if started_waiting == 0:
                    params.started_waiting = time.time()
                elif time.time() - started_waiting > waiting_period:
                    pick_up_fork = False
            elif len(forks) > 2:
                philosopher.logger.critical('%s has more than 2 forks!!!',
                                            philosopher.name)
                pick_up_fork = False
        else:
            thinking = params.thinking or 0
            if not thinking:
                philosopher.logger.warning('%s thinking...', philosopher.name)
            params.thinking = thinking + 1
        # Take action
        if pick_up_fork:
            self.pickup_fork(philosopher)
        else:
            self.release_forks(philosopher)
        
    def pickup_fork(self, philosopher):
        '''The philosopher has less than two forks. Check if forks are
available.'''
        right_fork = philosopher.params.number
        return philosopher.send(philosopher.monitor, 'pickup_fork', right_fork)\
                          .add_callback_args(self._continue, philosopher)
    
    def release_forks(self, philosopher):
        forks = philosopher.params.forks
        philosopher.params.forks = []
        philosopher.params.started_waiting = 0
        for fork in forks:
            philosopher.logger.debug('Putting down fork %s', fork)
            philosopher.send('monitor', 'putdown_fork', fork)
        # once released all the forks wait for a moment
        time.sleep(self.cfg.waiting_period)
        self._continue(None, philosopher)
    
    def _continue(self, fork, philosopher):
        if fork:
            forks = philosopher.params.forks
            if fork in forks:
                philosopher.logger.error('Got fork %s. I already have it', fork)
            else:
                philosopher.logger.debug('Got fork %s.', fork)
                forks.append(fork)
        self.take_action(philosopher)
    
    def actorparams(self, monitor, params):
        number = len(monitor.managed_actors) + 1
        name = 'Philosopher %s' % number
        params.update({'name': name,
                       'number': number,
                       'forks': []})
        return params
    

if __name__ == '__main__':
    DiningPhilosophers().start()


'''Configuration utilities which provides pulsar with configuration parameters
which can be parsed from the command line. Parsing is implemented using
the python argparser_ standard library module.

Config
~~~~~~~~~~

.. autoclass:: Config
   :members:
   :member-order: bysource

SettingBase
~~~~~~~~~~~~~~

.. autoclass:: SettingBase
   :members:
   :member-order: bysource
      
Setting
~~~~~~~~~~

.. autoclass:: Setting
   :members:
   :member-order: bysource


.. _argparser: http://docs.python.org/dev/library/argparse.html
'''
import copy
import inspect
import argparse
import os
import textwrap
import types
import logging

from pulsar import __version__, SERVER_NAME, parse_address
from . import system
from .httpurl import to_bytes, iteritems, HttpParser as PyHttpParser,\
                     native_str
from .importer import import_system_file


__all__ = ['Config',
           'SimpleSetting',
           'Setting',
           #'defaults',
           'ordered_settings',
           'validate_string',
           'validate_callable',
           'validate_bool',
           'validate_list',
           'validate_pos_int',
           'validate_pos_float',
           'make_optparse_options']

LOGGER = logging.getLogger('pulsar.config')

class DefaultSettings:

    def __init__(self):
        # port for serving a socket
        self.PORT = 8060
        # timeout for asynchronous Input/Output
        self.IO_TIMEOUT = None
        # Maximum number of concurrenct clients
        self.BACKLOG = 2048
        # Actors timeout
        self.TIMEOUT = 30
        # mailbox clients blocking
        self.message_timeout = 3

defaults = DefaultSettings()
section_docs = {}
KNOWN_SETTINGS = {}
KNOWN_SETTINGS_ORDER = []


def pass_through(arg):
    '''A dummy function accepting on parameter only. It does nothing. It is
used as default by :ref:`Application Hooks <setting-section-application-hooks>`.'''
    pass

def wrap_method(func):
    def _wrapped(instance, *args, **kwargs):
        return func(*args, **kwargs)
    return _wrapped

def ordered_settings():
    for name in KNOWN_SETTINGS_ORDER:
        yield KNOWN_SETTINGS[name]


class Config(object):
    '''A dictionary-like container of :class:`Setting` parameters for
fine tuning pulsar servers. It provides easy access to :attr:`Setting.value`
attribute by exposing the :attr:`Setting.name` as attribute.

:param description: description used when parsing the command line, same usage
    as in the :class:`argparse.ArgumentParser` class.
:param epilog: epilog used when parsing the command line, same usage
    as in the :class:`argparse.ArgumentParser` class.
:param version: version used when parsing the command line, same usage
    as in the :class:`argparse.ArgumentParser` class.
:param apps: list of application namespaces to include in the :attr:`settings`
    attribute.

.. attribute:: settings

    Dictionary of all :class:`Setting` instances available in this
    :class:`Config` container. Keys are given by the :attr:`SettingBase.name`
    attribute.
    
.. attribute:: params

    Dictionary of additional parameters which cannot be parsed in the
    command line.
'''
    exclude_from_config = set(('config',))
    
    def __init__(self, description=None, epilog=None,
                 version=None, apps=None, include=None,
                 exclude=None, settings=None, prefix=None,
                 name=None, **params):
        if settings is None:
            self.settings = {}
        else:
            self.settings = settings
        self.name = name
        self.prefix = prefix
        self.include = set(include or ())
        self.exclude = set(exclude or ())
        self.apps = set(apps or ())
        if settings is None:
            self.update_settings()
        self.params = {}
        self.description = description or 'Pulsar server'
        self.epilog = epilog or 'Have fun!'
        self.version = version or __version__
        self.update(params)
        
    def __iter__(self):
        return iter(self.settings)
    
    def __len__(self):
        return len(self.settings)
    
    def __contains__(self, name):
        return name in self.settings
    
    def items(self):
        for k, setting in iteritems(self.settings):
            yield k, setting.value
    
    def __getstate__(self):
        return self.__dict__.copy()

    def __setstate__(self, state):
        for k, v in state.items():
            self.__dict__[k] = v
        config = getattr(self, 'config', None)
        if config:
            self.import_from_module(config)

    def __getattr__(self, name):
        return self.get(name)

    def __setattr__(self, name, value):
        if name != "settings" and name in self.settings:
            raise AttributeError("Invalid access!")
        super(Config, self).__setattr__(name, value)

    def update(self, data):
        '''Update this :attr:`Config` with ``data`` which is either an
instance of Mapping or :class:`Config`.'''
        for name, value in data.items():
            if value is not None:
                self.set(name, value)
            
    def get(self, name, default=None):
        '''Get the value at ``name`` for this :class:`Config` container
following this algorithm:

* check ``name`` in the :attr:`settings` dictionary.
* if ``name`` is a known setting parameter return ``default``
* check ``name`` in the :attr:`params` dictionary
* raise :class:`AttributeError`.
'''
        if name not in self.settings:
            if name in KNOWN_SETTINGS:
                return default
            if name in self.params:
                return self.params[name]
            raise AttributeError("No configuration setting for: %s" % name)
        return self.settings[name].get()

    def set(self, name, value, default=False):
        '''Set the configuration :class:`Setting` at *name* with a new
*value*. If the *name* is not in this container, an :class:`AttributeError`
is raised. If ``default`` is ``True``, the :attr:`Setting.default` is
also set.'''
        if name not in self.settings:
            if self.prefix:
                prefix_name = '%s_%s' % (self.prefix, name)
                if prefix_name in self.settings:
                    return # dont set this value
            self.params[name] = value
        else:
            self.settings[name].set(value, default=default)

    def parser(self):
        '''Create the argparser_ for this configuration by adding all
settings via the :meth:`Setting.add_argument` method.

:rtype: an instance of :class:`argparse.ArgumentParser`.
'''
        kwargs = {
            "description": self.description,
            "epilog": self.epilog
        }
        parser = argparse.ArgumentParser(**kwargs)
        parser.add_argument('--version',
                            action='version',
                            version=self.version)
        return self.add_to_parser(parser)

    def add_to_parser(self, parser):
        setts = self.settings
        sorter = lambda x: (setts[x].section, setts[x].order)
        for k in sorted(setts, key=sorter):
            setts[k].add_argument(parser)
        return parser
        
    def import_from_module(self, mod=None):
        if mod:
            self.set('config', mod)
        try:
            mod = import_system_file(self.config)
        except Exception as e:
            raise RuntimeError('Failed to read config file "%s". %s' %\
                               (self.config, e))
        unknowns = []
        if mod:
            for k in dir(mod):
                # Skip private functions and attributes
                kl = k.lower()
                if k.startswith('_') or kl in self.exclude_from_config:
                    continue
                val = getattr(mod, k)
                # add unknown names to list
                if kl not in self.settings:
                    unknowns.append((k, val))
                else:
                    self.set(kl, val)
        return unknowns
        
    def on_start(self):
        '''Invoked by a :class:`pulsar.Application` just before starting.'''
        for sett in self.settings.values():
            sett.on_start()

    @property
    def workers(self):
        return self.settings['workers'].get()

    @property
    def address(self):
        '''An address to bind to, only available if a
:ref:`bind <setting-bind>` setting has been added to this :class:`Config`
container.'''
        bind = self.settings.get('bind')
        if bind:
            return parse_address(to_bytes(bind.get()))

    @property
    def uid(self):
        user = self.settings.get('user')
        if user:
            return system.get_uid(user.get())

    @property
    def gid(self):
        group = self.settings.get('group')
        if group:
            return system.get_gid(group.get())

    @property
    def proc_name(self):
        pn = self.settings.get('process_name')
        if pn:
            pn = pn.get()
        if pn is not None:
            return pn
        else:
            pn = self.settings.get('default_process_name')
            if pn:
                return pn.get()
        
    def copy(self, name=None, prefix=None):
        '''A deep copy of this :class:`Config` container. If ``prefix``
is given, it prefixes all non
:ref:`global settings <setting-section-global-server-settings>` with it.'''
        cls = self.__class__
        me = cls.__new__(cls)
        me.__dict__.update(self.__dict__)
        if prefix:
            me.prefix = prefix
        # Important, don't use the prefix from me.prefix!
        #        prefix = me.prefix
        settings = me.settings
        me.settings = {}
        for setting in settings.values():
            setting = setting.copy(name, prefix)
            me.settings[setting.name] = setting
        me.params = me.params.copy()
        return me
    
    def __copy__(self):
        return self.copy()
    
    def __deepcopy__(self, memo):
        return self.__copy__()
    
    ############################################################################
    ##    INTERNALS
    def update_settings(self):
        for s in ordered_settings():
            setting = s().copy(name=self.name, prefix=self.prefix)
            if setting.name in self.settings:  
                continue
            if setting.name not in self.include:
                if setting.name in self.exclude:
                    continue    # setting name in exclude set
                if setting.app and setting.app not in self.apps:
                    continue    # the setting is for an app not in the apps set
            self.settings[setting.name] = setting
    
            

class SettingMeta(type):
    '''A metaclass which collects all setting classes and put them
in the global ``KNOWN_SETTINGS`` list.'''
    def __new__(cls, name, bases, attrs):
        super_new = super(SettingMeta, cls).__new__
        parents = [b for b in bases if isinstance(b, SettingMeta)]
        val = attrs.get("validator")
        attrs["validator"] = wrap_method(val) if val else None
        if attrs.pop('virtual', False):
            return super_new(cls, name, bases, attrs)
        attrs["order"] = len(KNOWN_SETTINGS) + 1
        new_class = super_new(cls, name, bases, attrs)
        # build one instance to increase count
        new_class()
        new_class.fmt_desc(attrs['desc'] or '')
        if not new_class.name:
            new_class.name = new_class.__name__.lower()
        if new_class.name in KNOWN_SETTINGS_ORDER:
            old_class = KNOWN_SETTINGS.pop(new_class.name)
            new_class.order = old_class.order
        else:
            KNOWN_SETTINGS_ORDER.append(new_class.name)
        KNOWN_SETTINGS[new_class.name] = new_class
        return new_class

    def fmt_desc(cls, desc):
        desc = textwrap.dedent(desc).strip()
        setattr(cls, "desc", desc)
        lines = desc.split('\n\n')
        setattr(cls, "short", '' if not lines else lines[0])


class SettingBase(object):
    '''This is the base class of :class:`Settings` and :class:`SimpleSetting`.
It defines attributes for a given setting parameter.'''
    name = None
    '''The unique name used to access this setting in the :class:`Config`
    container.'''
    validator = None
    '''A validating function for this setting. It provided it must be a
function accepting one positional argument, the value to validate.'''
    value = None
    '''The actual value for this setting.'''
    default = None
    '''The default value for this setting.'''
    def __str__(self):
        return '{0} ({1})'.format(self.name, self.value)
    __repr__ = __str__

    def set(self, val, default=False):
        '''Set *val* as the :attr:`value` for this :class:`SettingBase`.
If *default* is ``True`` set also the :attr:`default` value.'''
        if hasattr(self.validator, '__call__'):
            val = self.validator(val)
        self.value = val
        if default:
            self.default = val

    def get(self):
        '''Returns :attr:`value`'''
        return self.value

    def on_start(self):
        '''Called when pulsar server starts. It can be used to perform
custom initialization for this :class:`Setting`.'''
        pass
    
    def copy(self, name=None, prefix=None):
        '''Copy this :class:`SettingBase`'''
        setting = copy.copy(self)
        if prefix and not setting.is_global:
            flags = setting.flags
            if flags and flags[-1].startswith('--'):
                # Prefix a setting
                setting.orig_name = setting.name 
                setting.name = '%s_%s' % (prefix, setting.name)
                setting.flags = ['--%s-%s' % (prefix, flags[-1][2:])]
        if name and not setting.is_global:
            setting.short = '%s application. %s' % (name , setting.short)
        return setting
    

class SimpleSetting(SettingBase):

    def __init__(self, name, value):
        self.name = name
        self.value = value
        self.orig_name = self.name


# This works for Python 2 and Python 3
class Setting(SettingMeta('BaseSettings', (SettingBase,), {'virtual': True})):
    '''A configuration parameter for pulsar. Parameters can be specified
on the command line or on a config file.'''
    creation_count = 0
    virtual = True
    '''If set to ``True`` the settings won't be loaded and it can be only used
as base class for other settings.'''
    nargs = None
    '''For positional arguments. Same usage as python :mod:`argparse` module.'''
    app = None
    '''Setting for a specific :class:`Application`.'''
    section = None
    '''Setting section, used for creating the
    :ref:`settings documentation <settings>`.'''
    flags = None
    '''List of options strings, e.g. ``[-f, --foo]``.'''
    choices = None
    '''Restrict the argument to the choices provided.'''
    type = None
    meta = None
    '''Same usage as ``metavar`` in the python :mod:`argparse` module. It is
    the name for the argument in usage message.'''
    action = None
    '''Same usage as ``action`` in the python
    :mod:`argparse.ArgumentParser.add_argument` method.'''
    short = None
    desc = None
    '''Description string'''
    is_global = False
    '''Flag used by pulsar :ref:`application framework <apps-framework>`
    when multiple application are used in a running server. If ``False``
    additional settings will be added to the :class:`Config` container.
    These settings are clones of this :class:`Setting` with
    name given by each application name and underacsore ``_``
    prefixing this :attr:`name`.'''
    orig_name = None

    def __init__(self, name=None, flags=None, action=None, type=None,
                 default=None, nargs=None, desc=None, validator=None,
                 app=None):
        self.default = default if default is not None else self.default
        if self.default is not None:
            self.set(self.default)
        self.name = name or self.name
        self.flags = flags or self.flags
        self.action = action or self.action
        self.nargs = nargs or self.nargs
        self.type = type or self.type
        self.desc = desc or self.desc
        self.app = app or self.app
        self.short = self.short or self.desc
        self.desc = self.desc or self.short
        if self.app and not self.section:
            self.section = self.app
        if not self.section:
            self.section = 'unknown'
        self.__class__.creation_count += 1
        if not hasattr(self, 'order'):
            self.order = 1000 + self.__class__.creation_count

    def __getstate__(self):
        return self.__dict__.copy()

    def add_argument(self, parser):
        '''Add this :class:`Setting` to the *parser*, an instance of
python :class:`argparse.ArgumentParser`, only if :attr:`flags` or
:attr:`nargs` and :attr:`name` are defined.'''
        kwargs = {'nargs':self.nargs}
        if self.type and self.type != 'string':
            kwargs["type"] = self.type
        if self.flags:
            args = tuple(self.flags)
            kwargs.update({"dest": self.name,
                           "action": self.action or "store",
                           "default": None,
                           "help": "%s [%s]" % (self.short, self.default)})
            if kwargs["action"] != "store":
                kwargs.pop("type",None)
                kwargs.pop("nargs",None)
        elif self.nargs and self.name:
            args = (self.name,)
            kwargs.update({'metavar': self.meta or None,
                           'help': self.short})
        else:
            # Not added to argparser
            return
        parser.add_argument(*args, **kwargs)
        

def validate_bool(val):
    if isinstance(val, bool):
        return val
    val = native_str(val)
    if not isinstance(val, str):
        raise TypeError("Invalid type for casting: %s" % val)
    if val.lower().strip() == "true":
        return True
    elif val.lower().strip() == "false":
        return False
    else:
        raise ValueError("Invalid boolean: %s" % val)

def validate_pos_int(val):
    if not isinstance(val, int):
        val = int(val, 0)
    else:
        # Booleans are ints!
        val = int(val)
    if val < 0:
        raise ValueError("Value must be positive: %s" % val)
    return val

def validate_pos_float(val):
    val = float(val)
    if val < 0:
        raise ValueError("Value must be positive: %s" % val)
    return val

def validate_string(val):
    va = native_str(val)
    if va is None:
        return None
    if not isinstance(va, str):
        raise TypeError("Not a string: %s" % val)
    return va.strip()

def validate_list(val):
    if val and not isinstance(val, (list, tuple)):
        raise TypeError("Not a list: %s" % val)
    return list(val)

def validate_dict(val):
    if val and not isinstance(val, dict):
        raise TypeError("Not a dictionary: %s" % val)
    return val

def validate_callable(arity):
    def _validate_callable(val):
        if not hasattr(val,'__call__'):
            raise TypeError("Value is not callable: %s" % val)
        if not inspect.isfunction(val):
            cval = val.__call__
            discount = 1
        else:
            discount = 0
            cval = val
        if arity != len(inspect.getargspec(cval)[0]) - discount:
            raise TypeError("Value must have an arity of: %s" % arity)
        return val
    return _validate_callable


def make_optparse_options(apps=None, exclude=None, include=None): # pragma nocover
    '''Create a tuple of optparse options'''
    from optparse import make_option
    class AddOptParser(list):
        def add_argument(self, *args, **kwargs):
            self.append(make_option(*args, **kwargs))
    config = Config(apps=apps, exclude=None, include=None)
    parser = AddOptParser()
    config.add_to_parser(parser)
    return tuple(parser)
    

################################################################################
##    Global Server Settings

section_docs['Global Server Settings'] = '''
These settings are global in the sense that they are used by the arbiter
as well as all pulsar workers. They are server configuration parameters.
'''
class Global(Setting):
    virtual = True
    section = "Global Server Settings"
    is_global = True


class ConfigFile(Global):
    name = "config"
    flags = ["-c", "--config"]
    meta = "FILE"
    validator = validate_string
    default = 'config.py'
    desc = """\
        The path to a Pulsar config file, where default Settings
        paramaters can be specified.
        """

    
class HttpProxyServer(Global):
    name = "http_proxy"
    flags = ["--http-proxy"]
    default = ''
    desc = """\
        The HTTP proxy server to use with HttpClient.
        """
    def on_start(self):
        if self.value:
            os.environ['http_proxy'] = self.value
            os.environ['https_proxy'] = self.value


class HttpParser(Global):
    name = "http_py_parser"
    flags = ["--http-py-parser"]
    action = "store_true"
    default = False
    desc = """\
    Set the python parser as default HTTP parser.
        """

    def on_start(self):
        if self.value:
            from pulsar.utils.httpurl import setDefaultHttpParser 
            setDefaultHttpParser(PyHttpParser)


class Debug(Global):
    name = "debug"
    flags = ["--debug"]
    validator = validate_bool
    action = "store_true"
    default = False
    desc = """\
        Turn on debugging in the server.

        This limits the number of worker processes to 1 and changes some error
        handling that's sent to clients.
        """


class Daemon(Global):
    name = "daemon"
    flags = ["-D", "--daemon"]
    validator = validate_bool
    action = "store_true"
    default = False
    desc = """\
        Daemonize the Pulsar process (posix only).

        Detaches the server from the controlling terminal and enters the
        background.
        """


class Pidfile(Global):
    name = "pidfile"
    flags = ["-p", "--pid"]
    meta = "FILE"
    validator = validate_string
    default = None
    desc = """\
        A filename to use for the PID file.

        If not set, no PID file will be written.
        """


class Password(Global):
    name = "password"
    flags = ["--password"]
    validator = validate_string
    default = None
    desc = """Set a password for the server"""


class User(Global):
    name = "user"
    flags = ["-u", "--user"]
    meta = "USER"
    validator = validate_string
    default = None
    desc = """\
        Switch worker processes to run as this user.

        A valid user id (as an integer) or the name of a user that can be
        retrieved with a call to pwd.getpwnam(value) or None to not change
        the worker process user.
        """


class Group(Global):
    name = "group"
    flags = ["-g", "--group"]
    meta = "GROUP"
    validator = validate_string
    default = None
    desc = """\
        Switch worker process to run as this group.

        A valid group id (as an integer) or the name of a user that can be
        retrieved with a call to pwd.getgrnam(value) or None to not change
        the worker processes group.
        """


class Loglevel(Global):
    name = "loglevel"
    flags = ["--log-level"]
    meta = "LEVEL"
    validator = validate_string
    default = "info"
    desc = """The granularity of log outputs.
            
            Valid level names are:
            
             * debug
             * info
             * warning
             * error
             * critical
             """

class LogHandlers(Global):
    name = "loghandlers"
    flags = ["--log-handlers"]
    default = ['console']
    validator = validate_list
    desc = """log handlers for pulsar server"""


class LogConfig(Global):
    name = "logconfig"
    default = {}
    validator = validate_dict
    desc = '''
    The logging configuration dictionary.

    This settings can only be specified on a config file
    '''


class Procname(Global):
    name = "process_name"
    flags = ["-n", "--name"]
    meta = "STRING"
    validator = validate_string
    default = None
    desc = """\
        A base to use with setproctitle for process naming.

        This affects things like ``ps`` and ``top``. If you're going to be
        running more than one instance of Pulsar you'll probably want to set a
        name to tell them apart. This requires that you install the setproctitle
        module.

        It defaults to 'pulsar'.
        """


class DefaultProcName(Global):
    name = "default_process_name"
    validator = validate_string
    default = SERVER_NAME
    desc = """\
        Internal setting that is adjusted for each type of application.
        """

################################################################################
##    Worker Processes

section_docs['Worker Processes'] = '''
This group of configuration parameters control the number of actors
for a given :class:`pulsar.Monitor`, the type of concurreny of the server and
other actor-specific parameters.
'''
    
class Workers(Setting):
    name = "workers"
    section = "Worker Processes"
    flags = ["-w", "--workers"]
    validator = validate_pos_int
    type = int
    default = 1
    desc = """\
        The number of workers for handling requests.
        
If using a multi-process concurrency, a number in the
the ``2-4 x NUM_CORES`` range should be good. If you are using threads this
number can be higher."""


class Concurrency(Setting):
    inherit = True # Inherited by the arbiter if an application is first created
    name = "concurrency"
    section = "Worker Processes"
    flags = ["--concurrency"]
    default = "process"
    desc = """\
        The type of concurrency to use: ``process`` or ``thread``.
        """


class MaxRequests(Setting):
    name = "max_requests"
    section = "Worker Processes"
    flags = ["--max-requests"]
    validator = validate_pos_int
    type = int
    default = 0
    desc = """\
        The maximum number of requests a worker will process before restarting.

        Any value greater than zero will limit the number of requests a worker
        will process before automatically restarting. This is a simple method
        to help limit the damage of memory leaks.

        If this is set to zero (the default) then the automatic worker
        restarts are disabled.
        """

class Backlog(Setting):
    name = "backlog"
    section = "Worker Processes"
    flags = ["--backlog"]
    validator = validate_pos_int
    type = int
    default = 2048
    desc = """\
        The maximum number of queued connections in a socket.
        
        This refers to the number of clients that can be waiting to be served.
        Exceeding this number results in the client getting an error when
        attempting to connect. It should only affect servers under significant
        load.
        Must be a positive integer. Generally set in the 64-2048 range.
        """


class Timeout(Setting):
    name = "timeout"
    section = "Worker Processes"
    flags = ["-t", "--timeout"]
    validator = validate_pos_int
    type = int
    default = 30
    desc = """\
        Workers silent for more than this many seconds are
        killed and restarted."""


################################################################################
##    APPLICATION HOOKS

section_docs['Application Hooks'] = '''
Application hooks are functions which can be specified in a
:ref:`config <setting-config>` file to perform custom tasks in a pulsar server.
These tasks can be scheduled when events occurs or at every event loop of
the various components of a pulsar application.

All application hooks are functions which accept one parameter only, the actor
invoking the function.'''
    
class Postfork(Setting):
    name = "post_fork"
    section = "Application Hooks"
    validator = validate_callable(1)
    type = "callable"
    default = staticmethod(pass_through)
    desc = """\
        Called just after a worker has been forked.

        The event loop is not yet available.
        """
        

class WhenReady(Setting):
    name = "when_ready"
    section = "Application Hooks"
    validator = validate_callable(1)
    type = "callable"
    default = staticmethod(pass_through)
    desc = """\
        Called just before a worker starts its event loop.
        
        This is a chance to setup :class:`pulsar.EventLoop` callbacks which
        can run periodically, at every loop or when some defined events occur.
        """


class ConnectionMade(Setting):
    name = "connection_made"
    section = "Application Hooks"
    validator = validate_callable(1)
    type = "callable"
    default = staticmethod(pass_through)
    desc = """\
        Called after a new connection is made.

        The callable needs to accept one instance variables for the
        connection instance.
        """


class ConnectionLost(Setting):
    name = "connection_lost"
    section = "Application Hooks"
    validator = validate_callable(1)
    type = "callable"
    default = staticmethod(pass_through)
    desc = """
        Called after a connection is lost.

        The callable needs to accept one instance variables for the
        connection instance.
        """


class PreRequest(Setting):
    name = "pre_request"
    section = "Application Hooks"
    validator = validate_callable(1)
    type = "callable"
    default = staticmethod(pass_through)
    desc = """\
        Called just before a worker processes the request.

        The callable needs to accept two instance variables for the Worker and
        the Request.
        """


class PostRequest(Setting):
    name = "post_request"
    section = "Application Hooks"
    validator = validate_callable(1)
    type = "callable"
    default = staticmethod(pass_through)
    desc = """\
        Called after a worker processes the request.

        The callable needs to accept two instance variables for the Worker and
        the Request.
        """



import sys
import signal
import ctypes
import ctypes.wintypes
import errno
import socket
import getpass
from time import sleep

from pulsar.utils.sockets import socket_pair

from .winprocess import WINEXE
from .base import *

__all__ = ['IOpoll',
           'close_on_exec',
           'Waker',
           'daemonize',
           'EXIT_SIGNALS',
           'get_uid',
           'get_gid',
           'get_maxfd']

# See: http://msdn.microsoft.com/en-us/library/ms724935(VS.85).aspx
SetHandleInformation = ctypes.windll.kernel32.SetHandleInformation
SetHandleInformation.argtypes = (ctypes.wintypes.HANDLE, ctypes.wintypes.DWORD,\
                                 ctypes.wintypes.DWORD)
SetHandleInformation.restype = ctypes.wintypes.BOOL

HANDLE_FLAG_INHERIT = 0x00000001

# The BREAK signal for windows
EXIT_SIGNALS = (signal.SIGINT, signal.SIGTERM, signal.SIGABRT, signal.SIGBREAK)
if sys.version_info >= (2, 7):
    SIG_NAMES[signal.CTRL_C_EVENT] = 'CTRL C EVENT'
    EXIT_SIGNALS += (signal.CTRL_C_EVENT,)
    
    
def get_parent_id():
    if ispy32:
        return os.getppid()
    else:
        return None

def chown(path, uid, gid):
    pass

def close_on_exec(fd):
    if fd:
        success = SetHandleInformation(fd, HANDLE_FLAG_INHERIT, 0)
        if not success:
            raise ctypes.GetLastError()
        
def _set_non_blocking(fd):
    pass

def get_uid(user=None):
    if not user:
        return getpass.getuser()
    elif user == getpass.getuser():
        return user

def get_gid(group=None):
    return None

def setpgrp():
    pass

def get_maxfd():
    return MAXFD

def daemonize():
    pass


class IOpoll(IOselect):
    '''The default windows IO poll class. Based on select.'''
    def poll(self, timeout=None):
        """Win32 select wrapper."""
        if not (self.read_fds or self.write_fds):
            # windows select() exits immediately when no sockets
            if timeout is None:
                timeout = 0.01
            else:
                timeout = min(timeout, 0.001)
            sleep(timeout)
            return ()
        # windows doesn't process 'signals' inside select(), so we set a max
        # time or ctrl-c will never be recognized
        if timeout is None or timeout > 0.5:
            timeout = 0.5
        return super(IOpoll, self).poll(timeout)
    
    
class Waker(object):
    '''In windows'''
    def __init__(self):
        self._writer, s = socket_pair(backlog=1)
        s.setblocking(True)   
        self._reader, addr = s.accept()
        self._writer.setblocking(False)
        self._reader.setblocking(False)
        s.close()
        
    def __str__(self):
        return 'Socket waker {0}'.format(self.fileno())
        
    def fileno(self):
        return self._reader.fileno()
    
    def wake(self):
        try:
            self._writer.send(b'x')
        except IOError:
            pass
        
    def consume(self):
        try:
            result = True
            while result:
                result = self._reader.recv(1024)
        except IOError:
            pass

    def close(self):
        self.reader.close()
        self.writer.close()
        



import sys
from copy import copy
from functools import partial

from pulsar import ProtocolError
from pulsar.utils.sockets import nice_address
from pulsar.async.defer import EventHandler, NOTHING

from .transport import TransportProxy, Connector, LOGGER


__all__ = ['Protocol', 'ProtocolConsumer', 'Connection', 'Producer']


class Protocol(EventHandler):
    '''Abstract class implemented in :class:`Connection`
and :class:`ProtocolConsumer`. It derives from :class:`EventHandler`
and therefore several events can be attached to subclasses.'''
    def connection_made(self, transport):
        '''Indicates that the :class:`Transport` is ready and connected
to the entity at the other end. The protocol should probably save the
transport reference as an instance variable (so it can call its write()
and other methods later), and may write an initial greeting or request
at this point.'''
        raise NotImplementedError
    
    def data_received(self, data):
        '''The :class:`Transport` has read some data from *other end*
and it invokes this method.'''
        raise NotImplementedError
    
    def eof_received(self):
        '''This is called when the other end called write_eof() (or
something equivalent).'''
        raise NotImplementedError
        
    def connection_lost(self, exc):
        '''The :class:`Transport` has been closed or aborted, has detected
that the other end has closed the connection cleanly, or has encountered an
unexpected error. In the first three cases the argument is None;
for an unexpected error, the argument is the exception that caused
the transport to give up.'''
        pass

    
class ProtocolConsumer(Protocol):
    '''The :class:`Protocol` consumer is one most important
:ref:`pulsar primitive <pulsar_primitives>`. It is responsible for receiving
incoming data from a the :meth:`Protocol.data_received` method implemented
in :class:`Connection`. It is used to decode and producing responses, i.e.
writing back to the client or server via
the :attr:`transport` attribute. The only method to implement should
be :meth:`Protocol.data_received`.

It has one :ref:`one time events <one-time-event>`:

* ``finish`` fired when this :class:`ProtocolConsumer` has finished consuming
  data and a response/exception is available. The :attr:`on_finished`
  is the :class:`Deferred` called back when this event occurs.

and one :ref:`many times events <many-times-event>`:

* ``data_received`` fired each time new data is received by this
  :class:`ProtocolConsumer` but not yet processed.
* ``data_processed`` fired each time new data is consumed by
  this :class:`ProtocolConsumer`.

.. attribute:: connection

    The :class:`Connection` of this consumer
    
.. attribute:: transport

    The :class:`Transport` of this consumer
    
.. attribute:: current_request

    Current request instance (used for clients only).
    
.. attribute:: connecting

    ``True`` if connecting to endpoint (for servers this is always false).
    
.. attribute:: producer

    The :class:`Producer` of this consumer.
    
.. attribute:: on_finished

    A :class:`Deferred` called once the :class:`ProtocolConsumer` has
    finished consuming protocol. It is called by the
    :attr:`connection` before disposing of this consumer. It is
    a proxy of ``self.event('finish')``.

'''
    ONE_TIME_EVENTS = ('finish',)
    MANY_TIMES_EVENTS = ('data_received', 'data_processed')
    def __init__(self, connection=None):
        super(ProtocolConsumer, self).__init__()
        self._connection = None
        self._current_request = None
        # this counter is updated by the connection
        self._data_received_count = 0
        # this counter is updated via the new_request method
        self._request_processed = 0
        # Number of times the consumer has tried to reconnect (for clients only)
        self._reconnect_retries = 0
        if connection is not None:
            connection.set_consumer(self)
    
    @property
    def connection(self):
        return self._connection
    
    @property
    def connected(self):
        return isinstance(self._connection, Connection)
    
    @property
    def connecting(self):
        return isinstance(self._connection, Connector)
    
    @property
    def event_loop(self):
        if self._connection:
            return self._connection.event_loop
    
    @property
    def current_request(self):
        return self._current_request
        
    @property
    def transport(self):
        if self._connection:
            return self._connection.transport
    
    @property
    def address(self):
        if self._connection:
            return self._connection.address
        
    @property
    def producer(self):
        if self._connection:
            return self._connection.producer
    
    @property
    def request_processed(self):
        '''The number of requests processed by this consumer.'''
        return self._request_processed
    
    @property
    def on_finished(self):
        return self.event('finish')
    
    def start_request(self):
        '''Invoked by the :meth:`new_request` method to kick start the
request with remote server/client. For server :class:`ProtocolConsumer` this
method is usually not implemented and therefore is simply a pass-through.
If implemented this is critical method
where errors caused by stale socket connections can arise.
**This method should not be called directly.** Use :meth:`new_request`
instead.'''
        pass
    
    def new_request(self, request=None):
        '''Starts a new *request* for this protocol consumer if
:attr:`connected` is `True`. There is no need to override this method,
implement :meth:`start_request` instead.'''
        if self.connected:
            self._request_processed += 1
            self._current_request = request
            self._connection.fire_event('pre_request', request)
            if request is not None:
                try:
                    self.start_request()
                except Exception:
                    self.finished(sys.exc_info())
        elif self.connecting:
            self._connection.add_callback(lambda r: self.new_request(request))
        else:
            raise RuntimeError('Cannot start new request. No connection.')
    
    def reset_connection(self):
        '''Cleanly dispose of the current :attr:`connection`. Used
by client consumers only.'''
        if self.connected:
            conn = self._connection
            clone = copy(self)
            conn._current_consumer = clone
            self._connection = None
            clone.finished()
            return clone
        
    def finished(self, result=NOTHING):
        '''Call this method when done with this :class:`ProtocolConsumer`.
By default it calls the :meth:`Connection.finished` method of the
:attr:`connection` attribute.'''
        if self._connection:
            return self._connection.finished(self, result)
        else:
            # No connection available. Just fire the finish callback
            self.fire_event('finish', result)
            return result
        
    def connection_lost(self, exc):
        return self.finished(exc)
        
    def can_reconnect(self, max_reconnect, exc):
        conn = self._connection
        # First we check if this was caused by a stale connection
        if conn and not self._data_received_count and conn.processed > 1:
            # switch off logging for this exception
            exc.logged = True
            return 1
        elif self._reconnect_retries < max_reconnect:
            self._reconnect_retries += 1
            exc.log()
            return self._reconnect_retries
        else:
            return 0
        
    def _data_received(self, data):
        # Called by Connection, it updates the counters and invoke
        # the high level data_received method which must be implemented
        # by subclasses
        self._data_received_count += 1 
        self._reconnect_retries = 0
        self.fire_event('data_received', data)
        result = self.data_received(data)
        self.fire_event('data_processed')
        return result
        
        
class Connection(Protocol, TransportProxy):
    '''A client or server connection with an end-point. This is not
connected until :meth:`Protocol.connection_made` is called by the
:class:`Transport`. This class is the bridge between the :class:`Transport`
and the :class:`ProtocolConsumer`. It has a :class:`Protocol`
interface and it routes data arriving from the :attr:`transport` to
the :attr:`current_consumer`, an instance of :class:`ProtocolConsumer`.

It has two :ref:`one time events <one-time-event>`:

* *connection_made*
* *connection_lost*

and three :ref:`many times events <many-times-event>`:

* *pre_request*
* *data_received*
* *post_request*

.. attribute:: producer

    The producer of this :class:`Connection`, It is either a :class:`Server`
    or a client :class:`Client`.
    
.. attribute:: transport

    The :class:`Transport` of this protocol connection. Initialised once the
    :meth:`Protocol.connection_made` is called.
    
.. attribute:: consumer_factory

    A factory of :class:`ProtocolConsumer` instances for this
    :class:`Connection`.
    
.. attribute:: session

    Connection session number. Created by the :attr:`producer`.
    
.. attribute:: processed

    Number of separate :class:`ProtocolConsumer` processed by this connection.
    
.. attribute:: current_consumer

    The :class:`ProtocolConsumer` currently handling incoming data.
'''
    ONE_TIME_EVENTS = ('connection_made', 'connection_lost')
    MANY_TIMES_EVENTS = ('data_received', 'pre_request', 'post_request')
    #
    def __init__(self, address, session, timeout, consumer_factory, producer):
        super(Connection, self).__init__()
        self._address = address
        self._session = session 
        self._processed = 0
        self._timeout = timeout
        self._idle_timeout = None
        self._current_consumer = None
        self._consumer_factory = consumer_factory
        self._producer = producer
        
    def __repr__(self):
        return '%s session %s' % (nice_address(self._address), self._session)
    
    def __str__(self):
        return self.__repr__()
    
    @property
    def session(self):
        return self._session
    
    @property
    def consumer_factory(self):
        return self._consumer_factory
    
    @property
    def current_consumer(self):
        return self._current_consumer
        
    @property
    def processed(self):
        return self._processed
    
    @property
    def address(self):
        return self._address
    
    @property
    def timeout(self):
        return self._timeout
    
    @property
    def producer(self):
        return self._producer
    
    def set_timeout(self, timeout):
        self._cancel_timeout()
        self._timeout = timeout
        self._add_idle_timeout()
        
    def set_consumer(self, consumer):
        '''Set a new :class:`ProtocolConsumer` for this :class:`Connection`.'''
        assert self._current_consumer is None, 'Consumer is not None'
        self._current_consumer = consumer
        consumer._connection = self
        self._processed += 1
    
    def connection_made(self, transport):
        # Implements protocol connection_made
        self._transport = transport
        # let everyone know we have a connection with endpoint
        self.fire_event('connection_made')
        self._add_idle_timeout()
        
    def data_received(self, data):
        self._cancel_timeout()
        while data:
            consumer = self._current_consumer
            if consumer is None:
                # New consumer.
                # These two lines are used by server connections only.
                consumer = self._consumer_factory(self)
                consumer.new_request()
            # Call the consumer _data_received method
            data = consumer._data_received(data)
            if data and self._current_consumer:
                # if data is returned from the response feed method and the
                # response has not done yet raise a Protocol Error
                raise ProtocolError('current consumer not done.')
        self._add_idle_timeout()
    
    def connection_lost(self, exc):
        '''Implements the :meth:`Protocol.connection_lost` method. It performs
these actions in the following order:

* Cancel the idle timeout if set.
* Fire the *connection_lost* :ref:`one time event <one-time-event>` with *exc*
  as event data.
* Invokes the connection_lost method in the :attr:`current_consumer` if
  available.'''
        if self.fire_event('connection_lost', exc):
            self._cancel_timeout()
            if self._current_consumer:
                self._current_consumer.connection_lost(exc)
                             
    def upgrade(self, consumer_factory):
        '''Update the :attr:`consumer_factory` attribute with a new
:class:`ProtocolConsumer` factory. This function can be used when the protocol
specification changes during a response (an example is a WebSocket
response).'''
        self._consumer_factory = consumer_factory
        
    def finished(self, consumer, result=NOTHING):
        '''Call this method to finish with the the current *consumer*.
the *consumer* must be the same as the :attr:`current_consumer` attribute.'''
        if consumer and consumer is self._current_consumer:
            # make sure the current consumer is set to None before callbacks
            self._current_consumer = None
            self.fire_event('post_request', consumer, sender=self)
            consumer.fire_event('finish', result, sender=self)
            consumer._connection = None
        else:
            raise RuntimeError()
    
    ############################################################################
    ##    INTERNALS
    def _timed_out(self):
        LOGGER.info('%s idle for %d seconds. Closing connection.',
                        self, self._timeout)
        self.close()
         
    def _add_idle_timeout(self):
        if not self.closed and not self._idle_timeout and self._timeout:
            self._idle_timeout = self.event_loop.call_later(self._timeout,
                                                            self._timed_out)
            
    def _cancel_timeout(self):
        if self._idle_timeout:
            self._idle_timeout.cancel()
            self._idle_timeout = None
         
         
class Producer(Protocol):
    '''A Producer of :class:`Connection` with remote servers or clients.
It is the base class for both :class:`Server` and :class:`ConnectionPool`.
The main method in this class is :meth:`new_connection` where a new
:class:`Connection` is created and added to the set of
:attr:`concurrent_connections`.

.. attribute:: connection_factory

    A factory producing the :class:`Connection` from a
    remote client with this producer.
    This attribute is used in the :meth:`new_connection` method.
    There shouldn't be any reason to change the default :class:`Connection`,
    it is here just in case.
    
.. attribute:: concurrent_connections

    Number of concurrent active connections
    
.. attribute:: received

    Total number of received connections
    
.. attribute:: timeout

    number of seconds to keep alive an idle connection
    
.. attribute:: max_connections

    Maximum number of connections allowed. A value of 0 (default)
    means no limit.
'''
    connection_factory = Connection
    def __init__(self, max_connections=0, timeout=0, connection_factory=None):
        super(Producer, self).__init__()
        self._received = 0
        self._timeout = timeout
        self._max_connections = max_connections
        self._concurrent_connections = set()
        if connection_factory:
            self.connection_factory = connection_factory
    
    @property
    def timeout(self):
        return self._timeout
    
    @property
    def received(self):
        return self._received
    
    @property
    def max_connections(self):
        return self._max_connections
    
    @property
    def concurrent_connections(self):
        return len(self._concurrent_connections)
    
    def new_connection(self, address, consumer_factory, producer=None):
        '''Called when a new :class:`Connection` is created. The *producer*
is either a :class:`Server` or a :class:`Client`. If the number of
:attr:`concurrent_connections` is greater or equal :attr:`max_connections`
a :class:`RuntimeError` is raised.'''
        if self._max_connections and self._received >= self._max_connections:
            raise RuntimeError('Too many connections')
        # increased the connections counter
        self._received = session = self._received + 1
        # new connection - not yet connected!
        producer = producer or self
        conn = self.connection_factory(address, session, self.timeout,
                                       consumer_factory, producer)
        # When the connection is made, add it to the set of
        # concurrent connections
        conn.bind_event('connection_made', self._add_connection)
        conn.copy_many_times_events(producer)
        conn.bind_event('connection_lost', partial(self._remove_connection,
                                                   conn))
        return conn
    
    def close_connections(self, connection=None, async=True):
        '''Close *connection* if specified, otherwise close all
active connections.'''
        if connection:
            connection.transport.close(async)
        else:
            for connection in list(self._concurrent_connections):
                connection.transport.close(async)
                
    def can_reuse_connection(self, connection, response):
        return True
            
    def _add_connection(self, connection):
        self._concurrent_connections.add(connection)
        
    def _remove_connection(self, connection, exc=None):
        # Called when the connection is lost
        self._concurrent_connections.discard(connection)
    